"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import uuid
from typing import Optional, List, Tuple, Iterable

from boto3 import client
from botocore.exceptions import ClientError

from syndicate.exceptions import InvalidValueError
from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry
from syndicate.core.constants import NONE_AUTH_TYPE, IAM_AUTH_TYPE
from syndicate.core.helper import dict_keys_to_capitalized_camel_case

_LOG = get_logger(__name__)

AUTH_TYPE_TO_STATEMENT_ID = {
    NONE_AUTH_TYPE: 'FunctionURLAllowPublicAccess-Syndicate',
    IAM_AUTH_TYPE: 'FunctionURLAllowIAMAccess-Syndicate'
}


def _str_list_to_list(param, param_name):
    if isinstance(param, list):
        result = param
    elif isinstance(param, Iterable):
        result = list(param)
    elif isinstance(param, str):
        result = [param]
    else:
        raise InvalidValueError(
            f"'{param_name}' must be a str or an iterable of strings.")
    return result


@apply_methods_decorator(retry())
class LambdaConnection(object):
    """ Lambda connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('lambda', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Lambda connection.')

    def create_lambda(self, lambda_name, func_name,
                      role, s3_bucket, s3_key, runtime='python3.10',
                      memory=128, timeout=300, architectures=None,
                      vpc_sub_nets=None, vpc_security_group=None,
                      env_vars=None, dl_target_arn=None, tracing_mode=None,
                      publish_version=False, layers=None,
                      ephemeral_storage=512, snap_start: str = None,
                      tags: dict = None):
        """ Create Lambda method
        :type lambda_name: str
        :type func_name: str
        :param func_name: name of the entry point function
        :param architectures: list function architecture type ['x86_64'|'arm64']
        :type role: str
        :param role: aws arn of role
        :type s3_bucket: str
        :type s3_key: str
        :type runtime: str
        :param memory: max limit of Lambda memory usage
        :param timeout: max limit of Lambda run time (300 - max value)
        :type vpc_sub_nets: list
        :type vpc_security_group: list
        :type env_vars: dict
        :param env_vars: {'string': 'string'}
        :type layers: list
        :param ephemeral_storage: amount of ephemeral storage between 512 MB
        and 10,240 MB
        :param snap_start: Optional[str] denotes `PublishedVersions`|`None`
        :type tags: dict
        :return: response
        """
        layers = [] if layers is None else layers
        params = dict(FunctionName=lambda_name, Runtime=runtime,
                      Role=role, Handler=func_name,
                      Code={'S3Bucket': s3_bucket, 'S3Key': s3_key},
                      Description=' ', Timeout=timeout, MemorySize=memory,
                      Publish=publish_version, Layers=layers,
                      EphemeralStorage={'Size': ephemeral_storage})
        if env_vars:
            params['Environment'] = {'Variables': env_vars}
        if vpc_sub_nets and vpc_security_group:
            params['VpcConfig'] = {
                'SubnetIds': vpc_sub_nets,
                'SecurityGroupIds': vpc_security_group
            }
        if dl_target_arn:
            params['DeadLetterConfig'] = {
                'TargetArn': dl_target_arn
            }
        if tracing_mode:
            params['TracingConfig'] = {
                'Mode': tracing_mode
            }
        if snap_start:
            params['SnapStart'] = {
                'ApplyOn': snap_start
            }
        if architectures:
            params['Architectures'] = architectures
        if tags:
            params['Tags'] = tags
        return self.client.create_function(**params)

    def get_existing_permissions(self, lambda_arn):
        try:
            # Retrieve the policy associated with the specified Lambda function
            policy = self.client.get_policy(FunctionName=lambda_arn)
            permissions = json.loads(policy['Policy']).get('Statement', [])

            return permissions
        except self.client.exceptions.ResourceNotFoundException:
            _LOG.debug(
                f'Can`t get permissions for lambda {lambda_arn}. '
                f'Lambda does not exist.'
            )
            return []

    def remove_permissions(self, lambda_arn, permissions_sids):
        for permission_sid in permissions_sids:
            try:
                self.client.remove_permission(
                    FunctionName=lambda_arn,
                    StatementId=permission_sid
                )
                _LOG.debug(f"Permissions deleted: {permission_sid},"
                           f" from lambda: {lambda_arn}")
            except self.client.exceptions.ClientError as e:
                _LOG.error(f"Can't delete permission: {permission_sid},"
                           f" from lambda: {lambda_arn}. Error: {e}")
                continue

    def set_url_config(self, function_name: str, qualifier: str = None,
                       auth_type: str = IAM_AUTH_TYPE, cors: dict = None,
                       principal: str = None, source_arn: str = None):
        _LOG.info(f'Setting url config for lambda: {function_name} with '
                  f'alias: {qualifier}')
        existing_url = self.get_url_config(function_name=function_name,
                                           qualifier=qualifier)
        if cors:
            # allow_origins is required for CORS
            if not cors.get('allow_origins'):
                cors['allow_origins'] = ['*']
        if not existing_url:
            _LOG.info('Existing url config was not found. Creating...')
            function_url = self.create_url_config(
                function_name=function_name, qualifier=qualifier,
                auth_type=auth_type, cors=cors)['FunctionUrl']
        else:
            _LOG.info('Existing url config was found. Updating...')
            existing_type = existing_url['AuthType']
            if existing_type != auth_type or existing_type == IAM_AUTH_TYPE:
                _LOG.warning('User has changed auth type or may have changed '
                             'principal or source arn. '
                             'Removing old permission')
                self.remove_one_permission(
                    function_name=function_name, qualifier=qualifier,
                    statement_id=AUTH_TYPE_TO_STATEMENT_ID[existing_type]
                )
            function_url = self.create_url_config(
                function_name=function_name, qualifier=qualifier,
                auth_type=auth_type, cors=cors, update=True)['FunctionUrl']

        if auth_type == NONE_AUTH_TYPE:
            _LOG.warning(f'Auth type is {NONE_AUTH_TYPE}. Setting '
                         f'the necessary resource-based policy')
            self.add_invocation_permission(
                name=function_name, principal='*', auth_type=auth_type,
                qualifier=qualifier, exists_ok=True,
                statement_id=AUTH_TYPE_TO_STATEMENT_ID[auth_type]
            )
        elif auth_type == IAM_AUTH_TYPE and principal:
            _LOG.warning(f'Auth type is {IAM_AUTH_TYPE}. Setting '
                         f'the necessary resource-based policy')
            self.add_invocation_permission(
                name=function_name, principal=principal,
                auth_type=auth_type, qualifier=qualifier,
                source_arn=source_arn,
                statement_id=AUTH_TYPE_TO_STATEMENT_ID[auth_type]
            )
        return function_url

    def put_function_event_invoke_config(self,
                                         function_name,
                                         max_retries=None):
        if max_retries is None:
            max_retries = 2
        return self.client.put_function_event_invoke_config(
            FunctionName=function_name,
            MaximumRetryAttempts=max_retries
        )

    def update_function_event_invoke_config(self,
                                            function_name,
                                            max_retries=None):
        if max_retries is None:
            max_retries = 2
        return self.client.update_function_event_invoke_config(
            FunctionName=function_name,
            MaximumRetryAttempts=max_retries
        )

    def delete_url_config(self, function_name: str, qualifier: str = None):
        params = dict(FunctionName=function_name)
        if qualifier:
            params['Qualifier'] = qualifier
        self.client.delete_function_url_config(**params)

    def create_url_config(self, function_name: str, qualifier: str = None,
                          auth_type: str = IAM_AUTH_TYPE, cors: dict = None,
                          update=False):
        params = dict(FunctionName=function_name,
                      AuthType=auth_type)
        if qualifier:
            params['Qualifier'] = qualifier
        if not cors:
            params['Cors'] = {}
        if cors and isinstance(cors, dict):
            params['Cors'] = dict_keys_to_capitalized_camel_case(cors)
        if update:
            return self.client.update_function_url_config(**params)
        else:
            return self.client.create_function_url_config(**params)

    def get_url_config(self, function_name: str, qualifier: str = None):
        params = dict(FunctionName=function_name)
        if qualifier:
            params['Qualifier'] = qualifier
        try:
            return self.client.get_function_url_config(**params)
        except ClientError as e:
            if e.response["Error"]["Code"] == 'ResourceNotFoundException':
                return None
            raise e

    def create_alias(self, function_name, name, version,
                     description=None, routing_config=None):
        """

        :param function_name: str
        :param name: str
        :param version: str
        :param description: str
        :param routing_config: dict str:int
        """
        params = dict(FunctionName=function_name, Name=name,
                      FunctionVersion=version)
        if description:
            params['Description'] = description
        if routing_config:
            params['RoutingConfig'] = routing_config

        return self.client.create_alias(**params)

    def get_aliases(self, function_name: str, name: str = None) -> dict:
        all_aliases = {}
        next_marker = 1  # to enter the loop
        while next_marker:
            req_param = {
                'FunctionName': function_name
            }
            if type(next_marker) == str:
                req_param['Marker'] = next_marker

            response = self.client.list_aliases(**req_param)

            all_aliases.update({item.get('Name'): item for item in
                                response.get('Aliases')})
            if all_aliases.get(name):
                return all_aliases.get(name)
            next_marker = response.get('NextMarker')
        return all_aliases

    def delete_alias(self, function_name: str, name: str) -> None:
        _LOG.debug(f'Removing lamba\'s \'{function_name}\' alias \'{name}\'')
        self.client.delete_alias(
            FunctionName=function_name,
            Name=name
        )

    def add_event_source(self, func_name, stream_arn, batch_size=10,
                         batch_window: Optional[int] = None,
                         start_position=None,
                         filters: Optional[List] = None,
                         function_response_types: Optional[List] = None):
        """ Create event source for Lambda
        :type func_name: str
        :type stream_arn: str
        :param batch_window: Optional[int]
        :param batch_size: max limit of Lambda event process in one time
        :param start_position: option for Lambda reading event mode
        :param filters: Optional[list]
        :param function_response_types: Optional[list] list of function
               response types
        :return: response
        """
        params = dict(
            EventSourceArn=stream_arn, FunctionName=func_name,
            Enabled=True, BatchSize=batch_size
        )
        if batch_window:
            params['MaximumBatchingWindowInSeconds'] = batch_window
        if start_position:
            params['StartingPosition'] = start_position
        if filters:
            params['FilterCriteria'] = {'Filters': filters}
        if function_response_types:
            params['FunctionResponseTypes'] = function_response_types
        response = self.client.create_event_source_mapping(**params)
        return response

    def list_event_sources(self, event_source_arn: Optional[str] = None,
                           function_name: Optional[str] = None) -> List:
        params = dict()
        if event_source_arn:
            params['EventSourceArn'] = event_source_arn
        if function_name:
            params['FunctionName'] = function_name
        return self.client.list_event_source_mappings(**params)['EventSourceMappings']

    def lambdas_list(self):
        """ Get all existing Lambdas.

        :return: list(if exists)
        """
        lambdas = []
        response = self.client.list_functions(MaxItems=1000)
        if 'Functions' in response:
            lambdas.extend(response['Functions'])
        marker = response.get('NextMarker')
        while marker:
            response = self.client.list_functions(Marker=marker,
                                                  MaxItems=1000)
            if 'Functions' in response:
                lambdas.extend(response['Functions'])
            marker = response.get('NextMarker')
        return lambdas

    def triggers_list(self, lambda_name):
        """ Get all existing triggers.

        :return: list(if exists)
        """
        mappings = []
        response = self.client.list_event_source_mappings(
            FunctionName=lambda_name)
        token = response.get('NextMarker')
        mappings.extend(response.get('EventSourceMappings'))
        while token:
            response = self.client.list_event_source_mappings(Marker=token)
            token = response.get('NextMarker')
            mappings.extend(response.get('EventSourceMappings'))
        return mappings

    def versions_list(self, function_name):
        versions = []
        response = self.client.list_versions_by_function(
            FunctionName=function_name)
        if 'Versions' in response:
            versions.extend(response['Versions'])
        marker = response.get('NextMarker')
        while marker:
            response = self.client.list_versions_by_function(Marker=marker)
            if 'Versions' in response:
                versions.extend(response['Versions'])
            marker = response.get('NextMarker')
        return versions

    def delete_lambda(self, func_name, log_not_found_error=True):
        """ Delete Lambda.

        :param func_name: str
        :param log_not_found_error: boolean, parameter is needed for proper log
        handling in the retry decorator
        """
        self.client.delete_function(FunctionName=func_name)

    def remove_event_sources(self, lambda_name):
        """ Remove trigger by name. Trigger has the same name as Lambda.

        :type lambda_name: str
        """
        triggers = self.triggers_list(lambda_name)
        for trigger in triggers:
            try:
                self.client.delete_event_source_mapping(
                    UUID=trigger['UUID'])
            except ClientError:
                _LOG.error('Failed to delete trigger.', exc_info=True)

    def remove_lambdas(self):
        """ Removes all specified lambdas.

        :type lambda_names: list
        """
        list_functions = self.lambdas_list()
        for each in list_functions:
            try:
                self.delete_lambda(each['FunctionName'])
            except ClientError:
                _LOG.error(
                    'Failed to delete lambda %s', each['FunctionName'],
                    exc_info=True)

    def remove_event_source(self, uuid):
        """ Delete event source stream.

        :param uuid: str
        """
        self.client.delete_event_source_mapping(UUID=uuid)

    def remove_invocation_permission(self, func_name, qualifier=None,
                                     ids_to_remove=None):
        """Removes permission for API Gateway to be able to invoke lambda
        :param func_name: the name/arn of the function to remove
        permissions from
        :type func_name: str
        :param qualifier: alias or version of the function
        :type qualifier: str
        :param ids_to_remove: specific ids of permissions to remove. If not
        specified, all the function's permissions will be removed
        :type ids_to_remove: list
        """
        ids_to_remove = ids_to_remove or []
        if not ids_to_remove:
            policies = self.get_policy(lambda_name=func_name)
            if not policies:
                return
            policies = json.loads(policies['Policy'])
            policies_meta = policies['Statement']
            ids_to_remove = []
            for policy in policies_meta:
                if policy['Action'] == 'lambda:InvokeFunction':
                    ids_to_remove.append(policy['Sid'])

        for sid in ids_to_remove:
            self.remove_one_permission(function_name=func_name,
                                       statement_id=sid,
                                       qualifier=qualifier)

    def remove_one_permission(self, function_name, statement_id=None,
                              qualifier=None, soft=True):
        params = dict(FunctionName=function_name, StatementId=statement_id)
        if qualifier:
            params['Qualifier'] = qualifier
        try:
            self.client.remove_permission(**params)
        except ClientError as e:
            if e.response["Error"]["Code"] == 'ResourceNotFoundException' \
                    and soft:
                return None
            raise e

    def add_invocation_permission(self, name, principal, source_arn=None,
                                  statement_id=None, auth_type=None,
                                  qualifier=None, exists_ok=False):
        """ Add permission for something to be able to invoke lambda
        :type name: str
        :type source_arn: str
        :type principal: str
        :type statement_id: str
        :type auth_type: str, NONE|AWS_IAM
        :type qualifier: str
        """
        action = 'lambda:InvokeFunctionUrl' if auth_type \
            else 'lambda:InvokeFunction'
        if not statement_id:
            statement_id = str(uuid.uuid1())
        params = dict(FunctionName=name, StatementId=statement_id,
                      Action=action, Principal=principal)
        if auth_type:
            params['FunctionUrlAuthType'] = auth_type
        if source_arn:
            params['SourceArn'] = source_arn
        if qualifier:
            params['Qualifier'] = qualifier
        try:
            return self.client.add_permission(**params)
        except ClientError as e:
            if e.response["Error"]["Code"] == 'ResourceConflictException' \
                    and exists_ok:
                return None
            raise e

    def update_code_source(self, lambda_name, s3_bucket, s3_key,
                           publish_version):
        """ Update code source (s3 bucket + file link) for specified lambda.

        :type lambda_name: str
        :type s3_bucket: str
        :type s3_key: str
        :type publish_version: bool
        """
        self.client.update_function_code(FunctionName=lambda_name,
                                         S3Bucket=s3_bucket,
                                         S3Key=s3_key,
                                         Publish=publish_version)

    def update_event_source(self, uuid, function_name, batch_size,
                            batch_window=None, filters: Optional[List] = None,
                            function_response_types: Optional[List] = None):
        params = dict(
            UUID=uuid, FunctionName=function_name, BatchSize=batch_size,
            Enabled=True
        )
        if batch_window is not None:
            params['MaximumBatchingWindowInSeconds'] = batch_window
        if filters is not None:
            params['FilterCriteria'] = {'Filters': filters}
        if function_response_types:
            params['FunctionResponseTypes'] = function_response_types
        else:
            params['FunctionResponseTypes'] = []
        return self.client.update_event_source_mapping(**params)

    def get_function(self, lambda_name, qualifier=None):
        """ Get function info if it is exists,
        else - ResourceNotFoundException.

        :type lambda_name: name
        :param qualifier: Using this optional parameter to specify a function
        version or an alias name.
        :type qualifier: str
        """
        params = dict(FunctionName=lambda_name)
        if qualifier:
            params['Qualifier'] = qualifier
        try:
            return self.client.get_function(**params)
        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                pass  # valid exception
            else:
                raise e

    def get_policy(self, lambda_name, qualifier=None):
        """ Returns the resource-based IAM policy for a function,
        version, or alias.

        :type lambda_name: name
        :param qualifier: Using this optional parameter to specify a function
        version or an alias name.
        :type qualifier: str
        """
        params = dict(FunctionName=lambda_name)
        if qualifier:
            params['Qualifier'] = qualifier
        try:
            return self.client.get_policy(**params)
        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                return  # valid exception
            else:
                raise e

    def invoke_lambda(self, lambda_name, invocation_type='Event',
                      log_type='Tail', client_context='', payload=b'',
                      qualifier=''):
        """

        :param lambda_name:
        :param invocation_type: 'Event' | 'RequestResponse' | 'DryRun'
        :param log_type: 'None' | 'Tail'
        :type client_context: str
        :type payload: b'bytes' | file
        :type qualifier: str
        :return:
        """
        params = dict(FunctionName=lambda_name,
                      InvocationType=invocation_type,
                      LogType=log_type,
                      ClientContext=client_context,
                      Payload=payload)
        if qualifier:
            params['Qualifier'] = qualifier
        return self.client.invoke(**params)

    def get_lambda_configuration(self, lambda_name, qualifier=None):
        params = dict(FunctionName=lambda_name)
        if qualifier:
            params['Qualifier'] = qualifier

        return self.client.get_function_configuration(**params)

    def update_lambda_configuration(self, lambda_name, role=None, handler=None,
                                    description=None, timeout=None,
                                    memory_size=None, vpc_sub_nets=None,
                                    vpc_security_group=None,
                                    env_vars=None, runtime=None,
                                    dead_letter_arn=None, kms_key_arn=None,
                                    layers=None, ephemeral_storage: int = None,
                                    snap_start: str = None,
                                    tracing_mode: str = None):
        params = dict(FunctionName=lambda_name)
        if ephemeral_storage:
            params['EphemeralStorage'] = {'Size': ephemeral_storage}
        if layers is not None:
            params['Layers'] = layers
        if role:
            params['Role'] = role
        if handler:
            params['Handler'] = handler
        if description:
            params['Description'] = description
        if timeout:
            params['Timeout'] = timeout
        if memory_size:
            params['MemorySize'] = memory_size
        if vpc_sub_nets is not None:
            params.setdefault('VpcConfig', {}).update({
                'SubnetIds': _str_list_to_list(vpc_sub_nets, 'VPC_SUB_NETS')
            })
        if vpc_security_group is not None:
            params.setdefault('VpcConfig', {}).update({
                'SecurityGroupIds': _str_list_to_list(vpc_security_group,
                                                      'VPC_SECURITY_GROUPS')
            })
        env_vars = env_vars or {}
        params['Environment'] = {'Variables': env_vars}
        if runtime:
            params['Runtime'] = runtime
        if dead_letter_arn:
            params['DeadLetterConfig'] = {'TargetArn': dead_letter_arn}
        if kms_key_arn:
            params['KMSKeyArn'] = kms_key_arn
        if snap_start:
            params['SnapStart'] = {
                'ApplyOn': snap_start
            }
        if not tracing_mode:
            params['TracingConfig'] = {
                'Mode': 'PassThrough'
            }
        else:
            params['TracingConfig'] = {
                'Mode': tracing_mode
            }
        return self.client.update_function_configuration(**params)

    def put_function_concurrency(self, function_name, concurrent_executions):
        return self.client.put_function_concurrency(
            FunctionName=function_name,
            ReservedConcurrentExecutions=concurrent_executions)

    def get_unresolved_concurrent_executions(self):
        return self.client.get_account_settings()['AccountLimit'][
            'UnreservedConcurrentExecutions']

    def publish_version(self, function_name, code_sha_256):
        return self.client.publish_version(
            FunctionName=function_name,
            CodeSha256=code_sha_256
        )

    def update_alias(self, function_name, alias_name, function_version):
        return self.client.update_alias(
            FunctionName=function_name,
            Name=alias_name,
            FunctionVersion=function_version
        )

    def create_layer(self, layer_name, s3_bucket, s3_key, runtimes,
                     description=None, architectures=None,
                     layer_license=None):
        kwargs = {'LayerName': layer_name, 'CompatibleRuntimes': runtimes,
                  'Content': {'S3Bucket': s3_bucket, 'S3Key': s3_key}}
        if description:
            kwargs['Description'] = description
        if layer_license:
            kwargs['LicenseInfo'] = layer_license
        if architectures:
            kwargs['CompatibleArchitectures'] = architectures
        return self.client.publish_layer_version(**kwargs)

    def get_lambda_layer_arn(self, name):
        lambda_layers = self.client.list_layers()
        for each in lambda_layers['Layers']:
            if each['LayerName'] == name:
                return each['LatestMatchingVersion']['LayerVersionArn']
        while lambda_layers.get('NextMarker'):
            lambda_layers = self.client.list_layers(
                Marker=lambda_layers.get('NextMarker'))
            for each in lambda_layers['Layers']:
                if each['LayerName'] == name:
                    return each['LatestMatchingVersion']['LayerVersionArn']

    def get_lambda_layer_by_arn(self, arn):
        return self.client.get_layer_version_by_arn(Arn=arn)

    def get_layer_version(self, name, version):
        return self.client.get_layer_version(LayerName=name,
                                             VersionNumber=version)

    def delete_layer(self, arn, log_not_found_error=True):
        """
        log_not_found_error parameter is needed for proper log handling in the
        retry decorator
        """
        version = arn.split(':')[len(arn.split(':')) - 1]
        arn = arn[:-len(version) - 1]
        return self.client.delete_layer_version(
            LayerName=arn,
            VersionNumber=int(version))

    def list_lambda_layer_versions(self, name, runtime=None):
        kwargs = {'LayerName': name}
        if runtime:
            kwargs['CompatibleRuntime'] = runtime
        response = self.client.list_layer_versions(**kwargs)
        versions = response['LayerVersions']

        while response.get('NextMarker'):
            kwargs['Marker'] = response['NextMarker']
            response = self.client.list_layer_versions(**kwargs)
            versions.extend(response['LayerVersions'])

        return versions

    def configure_provisioned_concurrency(self, name, qualifier,
                                          concurrent_executions):
        if type(concurrent_executions) is not int:
            raise InvalidValueError(
                f"Parameter 'concurrent_executions' "
                f"must be type of int, but not "
                f"'{type(concurrent_executions).__name__}'"
            )
        return self.client.put_provisioned_concurrency_config(
            FunctionName=name,
            Qualifier=qualifier,
            ProvisionedConcurrentExecutions=concurrent_executions
        )

    def get_provisioned_concurrency(self, name, qualifier):
        return self.client.get_provisioned_concurrency_config(
            FunctionName=name,
            Qualifier=qualifier
        )

    def describe_provisioned_concurrency_configs(self, name):
        configs = []
        response = self.client.list_provisioned_concurrency_configs(
            FunctionName=name
        )
        configs.extend(response.get('ProvisionedConcurrencyConfigs'))
        marker = response.get('NextMarker')
        while marker:
            response = self.client.list_provisioned_concurrency_configs(
                FunctionName=name,
                Marker=marker
            )
            configs.extend(response.get('ProvisionedConcurrencyConfigs'))
            marker = response.get('NextMarker')
        return configs

    def describe_function_concurrency(self, name):
        return self.client.get_function_concurrency(
            FunctionName=name
        ).get('ReservedConcurrentExecutions')

    def delete_function_concurrency_config(self, name):
        # client.delete_function_concurrency return 204 None in boto3 1.11.14
        self.client.delete_function_concurrency(
            FunctionName=name)

    def delete_provisioned_concurrency_config(self, name, qualifier):
        # client.delete_provisioned_concurrency_config return 204 None
        #   in boto3 1.11.14
        self.client.delete_provisioned_concurrency_config(
            FunctionName=name,
            Qualifier=qualifier
        )

    def list_function_versions(self, name):
        versions = []
        resp = self.client.list_versions_by_function(
            FunctionName=name,
            MaxItems=100
        )
        versions.extend(resp.get('Versions'))
        next_marker = resp.get('NextMarker')
        while next_marker:
            self.client.list_versions_by_function(
                FunctionName=name,
                MaxItems=100,
                Marker=next_marker
            )
            versions.extend(resp.get('Versions'))
            next_marker = resp.get('NextMarker')
        return versions

    def get_waiter(self, waiter_name):
        return self.client.get_waiter(waiter_name)

    def retrieve_vpc_config(self, response: dict) -> Tuple[set, set, Optional[str]]:
        """
        Retrieves subnets ids, security groups ids and vpc id from response
        received from lambda.get_function:
        response = {
            ...
            "VpcConfig": {
                "SubnetIds": [],
                "SecurityGroupIds": [],
                "VpcId": ""
            },
            ...
        }
        """
        _vpc = response.get('VpcConfig', {})
        _subnet_ids = set(_vpc.get('SubnetIds', []))
        _security_groups = set(_vpc.get('SecurityGroupIds', []))
        _vpc_id = _vpc.get('VpcId')
        return _subnet_ids, _security_groups, _vpc_id

    def retrieve_ephemeral_storage(self, response: dict) -> Optional[int]:
        """
        Works like the one above
        """
        return response.get('EphemeralStorage', {}).get('Size')



from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import WEB_SOCKET_API_GATEWAY_TYPE
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


class WebSocketApiGatewayGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = WEB_SOCKET_API_GATEWAY_TYPE
    CONFIGURATION = {
        'deploy_stage': None,
        'route_selection_expression': 'request.body.action',
        'resources': {
            "$connect": {
                "integration_type": "lambda",
                "enable_proxy": True,
                "lambda_alias": "${lambdas_alias_name}",
                "lambda_name": str
            },
            "$disconnect": {
                "integration_type": "lambda",
                "enable_proxy": True,
                "lambda_alias": "${lambdas_alias_name}",
                "lambda_name": str
            },
            "$default": {
                "integration_type": "lambda",
                "enable_proxy": True,
                "lambda_alias": "${lambdas_alias_name}",
                "lambda_name": str
            },
            "example": {
                "integration_type": "lambda",
                "enable_proxy": True,
                "lambda_alias": "${lambdas_alias_name}",
                "lambda_name": str
            }
        },
        'tags': dict
    }



import json

from syndicate.exceptions import ResourceNotFoundError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import DYNAMO_TABLE_TYPE
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator, BaseConfigurationGenerator
from syndicate.core.generators import (_read_content_from_file,
                                       _write_content_to_file)

_LOG = get_logger(
    'syndicate.core.generators.deployment_resources.dynamodb_generator')
USER_LOG = get_user_logger()


class DynamoDBGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = DYNAMO_TABLE_TYPE
    CONFIGURATION = {
        'hash_key_name': None,
        'hash_key_type': None,
        'sort_key_name': None,
        'sort_key_type': None,
        'billing_mode': str,
        'read_capacity': 1,
        'write_capacity': 1,
        'global_indexes': list,
        'autoscaling': list,
        'tags': dict
    }


class DynamoDBGlobalIndexGenerator(BaseConfigurationGenerator):
    CONFIGURATION = {
        'name': None,
        'index_key_name': None,
        'index_key_type': None,
        'index_sort_key_name': None,
        'index_sort_key_type': None
    }

    def __init__(self, **kwargs):
        self.table_name = kwargs.pop('table_name')
        super().__init__(**kwargs)

    def write(self):
        """Adds global index to dynamodb"""
        paths_with_table = self._get_resource_meta_paths(self.table_name,
                                                         DYNAMO_TABLE_TYPE)
        if not paths_with_table:
            raise ResourceNotFoundError(
                f"Table '{self.table_name}' was not found"
            )
        path_with_table = paths_with_table[0]  # table can be declared once
        USER_LOG.info(f"Adding global index to table '{self.table_name}'...")
        deployment_resources = json.loads(_read_content_from_file(
            path_with_table
        ))
        deployment_resources[self.table_name]['global_indexes'].append(
            self._resolve_configuration()
        )
        _write_content_to_file(path_with_table,
                               json.dumps(deployment_resources, indent=2))


class DynamoDBAutoscalingGenerator(BaseConfigurationGenerator):
    CONFIGURATION = {
        'resource_name': None,
        'role_name': 'AWSServiceRoleForApplicationAutoScaling_DynamoDBTable',
        'min_capacity': 1,
        'max_capacity': 10,
        'config': {
            "policy_name": None,
            "target_utilization": 70,
            "scale_in_cooldown": 60,  # seconds
            "scale_out_cooldown": 60
        },
        'dimension': "dynamodb:table:ReadCapacityUnits",
    }

    def __init__(self, **kwargs):
        self.table_name = kwargs.pop('table_name')
        kwargs['resource_name'] = self.table_name
        super().__init__(**kwargs)

    def write(self):
        """Adds autoscaling to dynamodb"""
        paths_with_table = self._get_resource_meta_paths(self.table_name,
                                                         DYNAMO_TABLE_TYPE)
        if not paths_with_table:
            raise ResourceNotFoundError(
                f"Table '{self.table_name}' was not found"
            )
        path_with_table = paths_with_table[0]
        USER_LOG.info(f"Adding autoscaling to table '{self.table_name}'...")
        deployment_resources = json.loads(_read_content_from_file(
            path_with_table
        ))
        deployment_resources[self.table_name]['autoscaling'].append(
            self._resolve_configuration()
        )
        _write_content_to_file(path_with_table,
                               json.dumps(deployment_resources, indent=2))



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import s3

from syndicate.exceptions import InvalidTypeError
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import to_logic_name, s3_bucket_logic_name

CANNED_ACL_PARAMS_MAPPING = {
    'private': s3.Private,
    'public-read': s3.PublicRead,
    'public-read-write': s3.PublicReadWrite,
    'aws-exec-read': 'AwsExecRead',
    'authenticated-read': s3.AuthenticatedRead,
    'bucket-owner-read': s3.BucketOwnerRead,
    'bucket-owner-full-control': s3.BucketOwnerFullControl,
    'log-delivery-write': s3.LogDeliveryWrite,
}


class CfS3Converter(CfResourceConverter):

    def convert(self, name, meta):
        bucket = s3.Bucket(s3_bucket_logic_name(name))
        bucket.BucketName = name
        self.template.add_resource(bucket)

        # location = meta.get('location')
        #   Ignore location as CloudFormation creates an Amazon S3 bucket in
        #   the same AWS Region where you create the AWS CloudFormation stack.

        acl = meta.get('acl')
        if acl:
            bucket.AccessControl = CANNED_ACL_PARAMS_MAPPING[acl]

        policy = meta.get('policy')
        if policy:
            bucket_policy = s3.BucketPolicy(
                to_logic_name('S3BucketPolicy', bucket.title))
            bucket_policy.Bucket = bucket.ref()
            bucket_policy.PolicyDocument = policy
            self.template.add_resource(bucket_policy)

        rule_document = meta.get('LifecycleConfiguration')
        if rule_document:
            rules = []
            for rule in rule_document['Rules']:
                if 'Prefix' not in rule:
                    rule['Prefix'] = ''
                rule_id = rule.pop('ID', None)
                if rule_id:
                    rule['Id'] = rule_id
                expiration = rule.pop('Expiration', None)
                if expiration:
                    if expiration.get('Date'):
                        rule['ExpirationDate'] = expiration.get('Date')
                    if expiration.get('Days'):
                        rule['ExpirationInDays'] = expiration.get('Days')
                    delete_marker = expiration.get('ExpiredObjectDeleteMarker')
                    if delete_marker:
                        rule['ExpiredObjectDeleteMarker'] = delete_marker
                transitions = rule.get('Transitions')
                if transitions:
                    for each in transitions:
                        transition_date = each.pop('Date', None)
                        transition_days = each.pop('Days', None)
                        if transition_date:
                            each['TransitionDate'] = transition_date
                        if transition_days:
                            each['TransitionInDays'] = transition_days
                rules.append(s3.LifecycleRule.from_dict(title=None, d=rule))
            bucket.LifecycleConfiguration = \
                s3.LifecycleConfiguration(Rules=rules)

        cors_configuration = meta.get('cors')
        if cors_configuration:
            cors_rules = []
            for rule in cors_configuration:
                # converting a rule to the expected format
                for key in list(rule.keys()):
                    new_key = key
                    if key == 'ID':
                        rule['Id'] = rule.pop(key)
                        continue
                    elif key == 'ExposeHeaders':
                        new_key = 'ExposedHeaders'
                        rule[new_key] = rule.pop(key)
                    elif key == 'MaxAgeSeconds':
                        new_key = 'MaxAge'
                        rule[new_key] = rule.pop(key)
                    if isinstance(rule[new_key], list) \
                            or isinstance(rule[new_key], int):
                        pass  # expected
                    elif isinstance(rule[new_key], str):
                        rule[new_key] = [rule[new_key]]
                    else:
                        raise InvalidTypeError(
                            f"CORS rule attribute '{key}' has invalid value: "
                            f"'{rule[key]}'. Should be str, int or list"
                        )
                cors_rules.append(s3.CorsRules.from_dict(title=None, d=rule))
            bucket.CorsConfiguration = \
                s3.CorsConfiguration(CorsRules=cors_rules)

    @staticmethod
    def configure_event_source_for_lambda(bucket, lambda_arn, events,
                                          filter_rules=None):
        try:
            notification_config = bucket.NotificationConfiguration
        except AttributeError:
            notification_config = s3.NotificationConfiguration()
            bucket.NotificationConfiguration = notification_config
        configs = []
        for event in events:
            config = s3.LambdaConfigurations(Event=event,
                                             Function=lambda_arn)
            configs.append(config)
            if filter_rules:
                rules = []
                for rule in filter_rules:
                    rules.append(s3.Rules(Name=rule['Name'],
                                          Value=rule['Value']))
                s3_key = s3.S3Key(Rules=rules)
                config.Filter = s3.Filter(S3Key=s3_key)
        try:
            existing_configs = notification_config.LambdaConfigurations
            existing_configs.extend(configs)
        except AttributeError:
            notification_config.LambdaConfigurations = configs



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import uuid

from syndicate.core.resources.lambda_resource import CLOUD_WATCH_RULE_TRIGGER, \
    EVENT_BRIDGE_RULE_TRIGGER
from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_helper import deploy_regions
from syndicate.core.transform.terraform.tf_resource_name_builder import \
    build_terraform_resource_name
from syndicate.core.transform.terraform.tf_resource_reference_builder import \
    build_cloud_watch_event_rule_name_ref, build_sns_topic_arn_ref


class SNSTopicConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        regions = deploy_regions(resource_name=name, meta=resource)
        default_region = self.config.region
        for region in regions:
            self.template.add_provider_if_not_exists(region=region)
            provider = None

            tf_resource_name = name
            if region != default_region:
                provider_type = self.template.provider_type()
                provider = f'{provider_type}.{region}'
                tf_resource_name = build_terraform_resource_name(name, region)

            self.create_sns_topic_in_region(name=name, resource=resource,
                                            provider=provider,
                                            tf_resource_name=tf_resource_name,
                                            region=region)

    def create_sns_topic_in_region(self, name, resource, tf_resource_name,
                                   region=None,
                                   provider=None):
        topic = sns_topic(topic_resource_name=tf_resource_name,
                          sns_topic_name=name, provider=provider)
        self.template.add_aws_sns_topic(meta=topic)
        self.process_event_sources(resource=resource,
                                   topic_tf_name=tf_resource_name,
                                   region=region)

    def process_event_sources(self, resource, topic_tf_name, region=None):
        event_sources = resource.get('event_sources')
        if event_sources:
            for event_source in event_sources:
                event_source_res_type = event_source.get('resource_type')
                if event_source_res_type in [CLOUD_WATCH_RULE_TRIGGER,
                                             EVENT_BRIDGE_RULE_TRIGGER]:
                    target_rule = event_source.get('target_rule')
                    target_rule = f'{target_rule}_{region}' if region else target_rule
                    rule_exp = build_cloud_watch_event_rule_name_ref(
                        target_rule=target_rule)

                    sns_topic_arn = build_sns_topic_arn_ref(
                        sns_topic=topic_tf_name)

                    trigger = cloud_watch_trigger(topic_tf_name=topic_tf_name,
                                                  rule_name=rule_exp,
                                                  topic_arn=sns_topic_arn)
                    self.template.add_aws_cloudwatch_event_target(trigger)
                    self._allow_service_invoke(topic=topic_tf_name)

    def _allow_service_invoke(self, topic):
        topic_policy_document = sns_topic_policy_document(topic=topic)
        document_json = json.dumps(topic_policy_document)
        topic_policy = sns_topic_policy(policy_json=document_json, topic=topic)
        self.template.add_aws_sns_topic_policy(meta=topic_policy)


def cloud_watch_trigger(topic_tf_name, topic_arn, rule_name):
    resource = {
        build_terraform_resource_name(topic_tf_name, 'trigger'):
            {
                "arn": topic_arn,
                "rule": rule_name
            }
    }
    return resource


def sns_topic_policy_document(topic):
    policy_document = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": str(uuid.uuid1()),
                "Effect": "Allow",
                "Principal":
                    {
                        "Service": 'events.amazonaws.com'
                    },
                "Action": "sns:Publish",
                "Resource": build_sns_topic_arn_ref(sns_topic=topic)
            }
        ]
    }
    return policy_document


def sns_topic_policy(policy_json, topic, provider=None):
    topic_policy_resource_name = build_terraform_resource_name(topic, 'policy')

    policy_meta = {
        "arn": build_sns_topic_arn_ref(sns_topic=topic),
        "policy": policy_json
    }
    if provider:
        policy_meta['provider'] = provider
    resource = {
        topic_policy_resource_name: policy_meta
    }
    return resource


def sns_topic(topic_resource_name, sns_topic_name, provider=None):
    topic = {
        "name": sns_topic_name
    }
    if provider:
        topic['provider'] = provider
    resource = {
        topic_resource_name: topic
    }
    return resource



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent
import glob
import json
import os
import shutil
import threading
from concurrent.futures.thread import ThreadPoolExecutor

from pathlib import Path

from syndicate.exceptions import EnvironmentError
from syndicate.commons.log_helper import get_logger
from syndicate.core.build.helper import build_py_package_name, zip_dir
from syndicate.core.conf.processor import path_resolver
from syndicate.core.constants import (LAMBDA_CONFIG_FILE_NAME,
                                      NODE_REQ_FILE_NAME,
                                      LAMBDA_LAYER_CONFIG_FILE_NAME,
                                      NODE_LAMBDA_LAYER_PATH, DEFAULT_SEP,
                                      LOCAL_REQ_FILE_NAME)
from syndicate.core.groups import RUNTIME_NODEJS
from syndicate.core.helper import (build_path, unpack_kwargs,
                                   execute_command_by_path, without_zip_ext,
                                   zip_ext)
from syndicate.core.project_state.project_state import BUILD_MAPPINGS
from syndicate.core.resources.helper import validate_params

_LOG = get_logger(__name__)


_JS_EXT = "*.js"
DEPENDENCIES_FOLDER = 'node_modules'


def _copy_js_files(search_path, destination_path):
    files = glob.iglob(build_path(search_path, _JS_EXT))
    for js_file in files:
        if os.path.isfile(js_file):
            shutil.copy2(js_file, destination_path)


def assemble_node_lambdas(project_path, bundles_dir, **kwargs):
    from syndicate.core import CONFIG
    project_abs_path = Path(CONFIG.project_path, project_path)
    _LOG.info(f'Going to package lambdas starting by path {project_abs_path}')
    _check_npm_is_installed()
    executor = ThreadPoolExecutor(max_workers=5)
    futures = []
    for root, sub_dirs, files in os.walk(project_abs_path):
        for item in files:
            if item.endswith(LAMBDA_CONFIG_FILE_NAME):
                _LOG.info(f'Going to build artifact in: {root}')
                arg = {
                    'item': item,
                    'root': root,
                    'target_folder': bundles_dir
                }
                futures.append(executor.submit(_build_node_artifact, arg))
            elif item.endswith(LAMBDA_LAYER_CONFIG_FILE_NAME):
                _LOG.info(f'Going to build lambda layer in `{root}`')
                arg = {
                    'layer_root': root,
                    'target_folder': bundles_dir
                }
                futures.append(executor.submit(build_node_lambda_layer, arg))
    for future in concurrent.futures.as_completed(futures):
        if future.result():
            _LOG.info(future.result())


@unpack_kwargs
def _build_node_artifact(item, root, target_folder):
    _LOG.debug(f'Building artifact in {target_folder}')
    lambda_config_dict = json.load(open(build_path(root, item)))
    _LOG.debug(f'Root path: {root}')
    req_params = ['lambda_path', 'name', 'version']
    validate_params(root, lambda_config_dict, req_params)
    lambda_name = lambda_config_dict['name']
    lambda_version = lambda_config_dict['version']
    artifact_name = lambda_name + '-' + lambda_version
    package_name = build_py_package_name(lambda_name, lambda_version)
    artifact_path = str(Path(target_folder, artifact_name))

    shutil.copytree(root, str(Path(artifact_path, 'lambdas', lambda_name)))
    install_requirements(root, target_folder, artifact_path, package_name)


@unpack_kwargs
def build_node_lambda_layer(layer_root: str, target_folder: str):
    with open(Path(layer_root, LAMBDA_LAYER_CONFIG_FILE_NAME), 'r') as file:
        layer_config = json.load(file)

    validate_params(layer_root, layer_config, ['name', 'deployment_package'])
    artifact_name = without_zip_ext(layer_config['deployment_package'])
    package_name = zip_ext(layer_config['deployment_package'])
    artifact_path = str(Path(target_folder, artifact_name))
    modules_path = str(Path(artifact_path, DEPENDENCIES_FOLDER))

    shutil.copytree(layer_root, modules_path)
    install_requirements(layer_root, target_folder, artifact_path,
                         package_name, is_layer=True)


def install_requirements(root: str, target_folder: str, artifact_path: str,
                         package_name: str, is_layer=False):
    """
    artifact_path: str - Absolute archive path
    root: str - lambda folder (src/lambdas/{$lambda_name})
    """
    _LOG.info(f'Artifacts path: {artifact_path}')
    os.makedirs(artifact_path, exist_ok=True)
    if not os.path.exists(artifact_path):
        os.makedirs(artifact_path)
    _LOG.debug('Folders are created')
    # getting file content
    req_path = Path(root, NODE_REQ_FILE_NAME)

    try:
        if os.path.exists(req_path):
            command = 'npm install'
            # this command creates 'node_modules' folder in lambda folder
            execute_command_by_path(command=command, path=root)
            _LOG.debug('3-rd party dependencies were installed successfully')
            try:
                shutil.copytree(Path(root, DEPENDENCIES_FOLDER),
                                Path(artifact_path, DEPENDENCIES_FOLDER),
                                dirs_exist_ok=True)
            except FileNotFoundError:
                _LOG.info('No dependencies folder - nothing to copy.')
            except Exception as e:
                _LOG.exception(f'Error occurred while lambda files coping: {e}')

        # install local requirements
        local_requirements_path = Path(root, LOCAL_REQ_FILE_NAME)
        if os.path.exists(local_requirements_path):
            _LOG.info('Going to install local dependencies')
            _copy_local_req(artifact_path, local_requirements_path)
            _LOG.info('Local dependencies were installed successfully')

        if is_layer:
            zip_dir(artifact_path,
                    str(build_path(target_folder, package_name)),
                    NODE_LAMBDA_LAYER_PATH)
        else:
            zip_dir(artifact_path,
                    build_path(target_folder, package_name))

        lock = threading.RLock()
        lock.acquire()
        try:
            # remove unused folder/files
            node_modules_path = os.path.join(root, DEPENDENCIES_FOLDER)
            if os.path.exists(node_modules_path):
                shutil.rmtree(node_modules_path)
            # todo Investigate deleting package_lock file
            # shutil.rmtree(os.path.join(root, 'package_lock.json'))
            shutil.rmtree(artifact_path)
        except FileNotFoundError as e:
            _LOG.exception(f'Error occurred while temp files removing: {e}')
        finally:
            lock.release()
        return f'Lambda package {package_name} was created successfully'
    except Exception as e:
        _msg = 'Error occurred during the lambda layer deployment package ' \
               'assembling'
        _LOG.exception(f'{_msg}: {e}')
        return _msg


def _check_npm_is_installed():
    import subprocess
    result = subprocess.call('npm -v', shell=True)
    if result:
        raise EnvironmentError(
            'NPM is not installed. There is no ability to build '
            'NodeJS bundle. Please, install npm and retry to build bundle.')


def _copy_local_req(artifact_path, local_req_path):
    from syndicate.core import CONFIG
    with open(local_req_path) as f:
        local_req_list = f.readlines()
    local_req_list = [path_resolver(r.strip()) for r in local_req_list]
    _LOG.info(f'Installing local dependencies: {local_req_list}')
    # copy folders
    for lrp in local_req_list:
        _LOG.info(f'Processing local dependency: {lrp}')
        shutil.copytree(str(Path(CONFIG.project_path,
                                 BUILD_MAPPINGS[RUNTIME_NODEJS], lrp)),
                        str(Path(artifact_path, lrp)))
        _LOG.debug('Dependency was copied successfully')

        folders = [r for r in lrp.split(DEFAULT_SEP) if r]
        # process folder from root project
        folders.insert(0, '')
        for folder in folders:
            src_path = Path(CONFIG.project_path,
                            BUILD_MAPPINGS[RUNTIME_NODEJS], folder)
            dst_path = Path(artifact_path, folder)
            _copy_js_files(str(src_path), str(dst_path))
        _LOG.debug('JavaScript files from packages were copied successfully')



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import sqs

from syndicate.exceptions import InvalidValueError
from syndicate.core.resources.sqs_resource import SqsResource
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import to_logic_name, sqs_queue_logic_name


class CfSqsConverter(CfResourceConverter):

    def convert(self, name, meta):
        is_fifo = bool(meta.get('fifo_queue', False))
        queue_name = SqsResource.build_resource_name(is_fifo, name)

        delay_sec = meta.get('delay_seconds')
        max_mes_size = meta.get('maximum_message_size')
        mes_ret_period = meta.get('message_retention_period')
        policy = meta.get('policy')
        receive_mes_wait_sec = meta.get('receive_message_wait_time_seconds')
        redrive_policy = meta.get('redrive_policy')
        vis_timeout = meta.get('visibility_timeout')
        kms_master_key_id = meta.get('kms_master_key_id')
        kms_data_reuse_period = meta.get('kms_data_key_reuse_period_seconds')
        content_deduplication = bool(meta.get('content_based_deduplication',
                                              False))

        queue = sqs.Queue(sqs_queue_logic_name(name))
        queue.QueueName = queue_name
        self.template.add_resource(queue)

        if delay_sec:
            if delay_sec < 0 or delay_sec > 900:
                raise InvalidValueError(
                    'Delay seconds for queue must be between 0 and 900 seconds'
                )
            queue.DelaySeconds = delay_sec

        if max_mes_size:
            if max_mes_size < 1024 or max_mes_size > 262144:
                raise InvalidValueError(
                    'Maximum message size must be between 1024 and 262144 '
                    'bytes'
                )
            queue.MaximumMessageSize = max_mes_size

        if mes_ret_period:
            if mes_ret_period < 60 or mes_ret_period > 1209600:
                raise InvalidValueError(
                    'Message retention size must be between 60 and 1209600 '
                    'seconds'
                )
            queue.MessageRetentionPeriod = mes_ret_period

        if receive_mes_wait_sec:
            if receive_mes_wait_sec < 0 or receive_mes_wait_sec > 20:
                raise InvalidValueError(
                    'Receive message wait time must be between 0 and 20 '
                    'seconds'
                )
            queue.ReceiveMessageWaitTimeSeconds = receive_mes_wait_sec

        if redrive_policy:
            queue.RedrivePolicy = sqs.RedrivePolicy(
                deadLetterTargetArn=redrive_policy['deadLetterTargetArn'],
                maxReceiveCount=redrive_policy['maxReceiveCount']
            )

        if vis_timeout:
            if vis_timeout < 0 or vis_timeout > 43200:
                raise InvalidValueError(
                    'Visibility timeout must be between 0 and 43200 seconds'
                )
            queue.VisibilityTimeout = vis_timeout

        if kms_master_key_id:
            queue.KmsMasterKeyId = kms_master_key_id

        if kms_data_reuse_period:
            if kms_data_reuse_period < 60 or kms_data_reuse_period > 86400:
                raise InvalidValueError(
                    'KMS key reuse period must be between 60 and 86400 seconds'
                )
            queue.KmsDataKeyReusePeriodSeconds = kms_data_reuse_period

        if is_fifo:
            queue.FifoQueue = True

        if content_deduplication:
            queue.ContentBasedDeduplication = True

        if policy:
            queue_policy = sqs.QueuePolicy(
                to_logic_name('SQSQueuePolicy', name))
            queue_policy.PolicyDocument = policy
            queue_policy.Queues = [queue.ref()]
            self.template.add_resource(queue_policy)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import batch

from syndicate.exceptions import ResourceNotFoundError
from syndicate.core.helper import dict_keys_to_upper_camel_case
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import to_logic_name, iam_role_logic_name


class CfBatchJobDefinitionConverter(CfResourceConverter):

    def convert(self, name, meta):
        container_props = meta.get('container_properties')
        if container_props:
            container_props = self.prepare_container_properties(
                container_props=container_props,
                job_definition_name=name)

        logic_name = to_logic_name('BatchJobDefinition', name)
        job_def = batch.JobDefinition(logic_name)
        job_def.JobDefinitionName = name
        job_def.Type = meta['job_definition_type']

        if meta.get('parameters'):
            job_def.Parameters = meta['parameters']

        if container_props:
            job_def.ContainerProperties = batch.ContainerProperties.from_dict(
                title=None, d=container_props)

        node_props = meta.get('node_properties')
        if node_props:
            container_props = \
                node_props.get('node_range_properties', {}).get('container')
            if container_props:
                container_props = self.prepare_container_properties(
                    container_props=container_props,
                    job_definition_name=name)
                del(node_props['node_range_properties']['container'])

            node_props = dict_keys_to_upper_camel_case(node_props)
            if container_props:
                node_props['node_range_properties']['container'] = \
                    container_props
            job_def.NodeProperties = batch.NodeProperties.from_dict(
                title=None, d=node_props)

        if meta.get('retry_strategy'):
            retry_strategy = dict_keys_to_upper_camel_case(
                meta['retry_strategy'])
            job_def.RetryStrategy = batch.RetryStrategy.from_dict(
                title=None, d=retry_strategy)

        if meta.get('propagate_tags') is not None:
            job_def.PropagateTags = meta['propagate_tags']
        if meta.get('timeout'):
            timeout = dict_keys_to_upper_camel_case(meta['timeout'])
            job_def.Timeout = batch.Timeout.from_dict(title=None, d=timeout)
        if meta.get('tags'):
            job_def.Tags = meta['tags']
        if meta.get('platform_capabilities'):
            job_def.PlatformCapabilities = meta['platform_capabilities']
        self.template.add_resource(job_def)

    def prepare_container_properties(self, container_props,
                                     job_definition_name):
        job_role_arn = container_props.get('job_role_arn')
        execution_role_arn = container_props.get('execution_role_arn')
        log_conf_options = container_props \
            .get('log_configuration', {}).get('options', {}).copy()
        container_props = dict_keys_to_upper_camel_case(container_props)
        if job_role_arn:
            container_props['JobRoleArn'] = \
                self.check_role_exists(job_role_arn, job_definition_name)
        if execution_role_arn:
            container_props['ExecutionRoleArn'] = \
                self.check_role_exists(execution_role_arn, job_definition_name)
        if log_conf_options:
            container_props['log_configuration']['options'] = \
                log_conf_options
        return container_props

    def check_role_exists(self, role_arn, job_definition_name):
        role = self.get_resource(iam_role_logic_name(role_arn))
        if role:
            return role.get_att('Arn')
        iam_conn = self.resources_provider.iam().iam_conn
        existing_role = iam_conn.check_if_role_exists(role_name=role_arn)
        if existing_role:
            return existing_role
        raise ResourceNotFoundError(
            f"IAM Role '{role_arn}' specified in '{job_definition_name}' "
            f"batch job definition does not exist"
        )



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import SNS_TOPIC_TYPE


class SNSTopicGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = SNS_TOPIC_TYPE
    CONFIGURATION = {
        'region': None,
        'event_sources': list,
        'tags': dict
    }



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import re
from datetime import datetime, timedelta, timezone

from botocore.exceptions import ClientError

from syndicate.exceptions import InternalError, ConfigurationError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.connection import ConnectionProvider
from syndicate.connection.sts_connection import STSConnection
from syndicate.core.conf.processor import ConfigHolder
from syndicate.core.project_state.project_state import ProjectState
from syndicate.core.resources.processors_mapping import ProcessorFacade
from syndicate.core.resources.resources_provider import ResourceProvider
from syndicate.core.conf.bucket_view import URIBucketView, RegexViewDigest, \
    NAMED_S3_URI_PATTERN, S3_PATTERN_GROUP_NAMES
from syndicate.core.helper import handle_interruption

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()

SESSION_TOKEN = 'aws_session_token'
SECRET_KEY = 'aws_secret_access_key'
ACCESS_KEY = 'aws_access_key_id'

CONFIRMATION_ANSWERS = ['y', 'yes']

# if temporary credentials will be expired in less than N minutes,
# syndicate will regenerate temp credentials instead of using old ones
CREDENTIALS_EXPIRATION_THRESHOLD_MIN = 15


def exception_handler(exception_type, exception, traceback):
    print(exception)


# sys.excepthook = exception_handler

# CONF VARS ===================================================================
CONF_PATH = os.environ.get('SDCT_CONF')
CONFIG: ConfigHolder = None
CONN = None
CREDENTIALS = None
RESOURCES_PROVIDER: ResourceProvider = None
PROCESSOR_FACADE: ProcessorFacade = None
PROJECT_STATE: ProjectState = None


def _ready_to_assume():
    return CONFIG.access_role and not CONFIG.use_temp_creds


def _ready_to_use_creds():
    return not CONFIG.access_role and CONFIG.aws_access_key_id and \
           CONFIG.aws_secret_access_key and not CONFIG.use_temp_creds


def _ready_to_use_provided_temp_creds():
    has_temporary_creds_set = CONFIG.temp_aws_access_key_id \
                              and CONFIG.temp_aws_secret_access_key \
                              and CONFIG.temp_aws_session_token \
                              and CONFIG.expiration

    credentials_expired = True
    
    if has_temporary_creds_set:
        credentials_expired = validate_temp_credentials_expiration(
            expiration=CONFIG.expiration
        )
        if credentials_expired and CONFIG.use_temp_creds:
            raise ConfigurationError('Temporary credentials have expired')

    return has_temporary_creds_set and not credentials_expired


def _ready_to_generate_temp_creds():
    return not CONFIG.access_role and CONFIG.use_temp_creds


def initialize_connection():
    global CONFIG
    global CONN
    global CONF_PATH
    global CREDENTIALS
    global RESOURCES_PROVIDER
    global PROCESSOR_FACADE

    regex_digest = RegexViewDigest()
    regex_digest.expression = NAMED_S3_URI_PATTERN
    regex_digest.groups = S3_PATTERN_GROUP_NAMES
    uri_bucket_view = URIBucketView()
    uri_bucket_view.digest = regex_digest

    CONFIG = ConfigHolder(CONF_PATH)
    CONFIG.deploy_target_bucket_view = uri_bucket_view
    sts = STSConnection(CONFIG.region, CONFIG.aws_access_key_id,
                        CONFIG.aws_secret_access_key, CONFIG.aws_session_token)
    try:
        CREDENTIALS = {
            'region': CONFIG.region
        }
        if _ready_to_use_provided_temp_creds():
            _LOG.debug(f'Going to use previously generated temporary '
                       f'credentials')
            CREDENTIALS[ACCESS_KEY] = CONFIG.temp_aws_access_key_id
            CREDENTIALS[SECRET_KEY] = CONFIG.temp_aws_secret_access_key
            CREDENTIALS[SESSION_TOKEN] = CONFIG.temp_aws_session_token
        elif _ready_to_assume():
            _LOG.debug('Starting to assume role ...')
            # get CREDENTIALS for N hours

            token_code = prompt_mfa_code() if CONFIG.serial_number else None
            temp_credentials = sts.get_temp_credentials(
                role_arn=CONFIG.access_role,
                acc_id=CONFIG.account_id,
                duration=CONFIG.session_duration,
                serial_number=CONFIG.serial_number,
                token_code=token_code
            )
            _LOG.debug(f'Role {CONFIG.access_role} is assumed successfully'
                       f'for {CONFIG.session_duration} seconds')
            CREDENTIALS[ACCESS_KEY] = temp_credentials['AccessKeyId']
            CREDENTIALS[SECRET_KEY] = temp_credentials['SecretAccessKey']
            CREDENTIALS[SESSION_TOKEN] = temp_credentials['SessionToken']
            _LOG.debug(f'Temporary credentials have been successfully '
                       f'generated by assuming the role {CONFIG.access_role}, '
                       f'saving to config.')
            CONFIG.set_temp_credentials_to_config(
                temp_aws_access_key_id=temp_credentials['AccessKeyId'],
                temp_aws_secret_access_key=temp_credentials['SecretAccessKey'],
                temp_aws_session_token=temp_credentials['SessionToken'],
                expiration=temp_credentials['Expiration']
            )
        elif _ready_to_generate_temp_creds():
            _LOG.debug(f'Going to generate new temporary credentials')

            token_code = None
            if CONFIG.serial_number:
                token_code = prompt_mfa_code()
            temp_credentials = sts.get_session_token(
                duration=CONFIG.session_duration,
                serial_number=CONFIG.serial_number,
                token_code=token_code
            )
            CREDENTIALS[ACCESS_KEY] = temp_credentials['AccessKeyId']
            CREDENTIALS[SECRET_KEY] = temp_credentials['SecretAccessKey']
            CREDENTIALS[SESSION_TOKEN] = temp_credentials['SessionToken']
            _LOG.debug(f'Temporary credentials have been successfully '
                       f'generated, saving to config.')
            CONFIG.set_temp_credentials_to_config(
                temp_aws_access_key_id=temp_credentials['AccessKeyId'],
                temp_aws_secret_access_key=temp_credentials['SecretAccessKey'],
                temp_aws_session_token=temp_credentials['SessionToken'],
                expiration=temp_credentials['Expiration']
            )
        elif _ready_to_use_creds():
            _LOG.debug('Credentials access')
            CREDENTIALS[ACCESS_KEY] = CONFIG.aws_access_key_id
            CREDENTIALS[SECRET_KEY] = CONFIG.aws_secret_access_key
        CONN = ConnectionProvider(CREDENTIALS)
        RESOURCES_PROVIDER = ResourceProvider(config=CONFIG,
                                              credentials=CREDENTIALS,
                                              sts_conn=sts)
        PROCESSOR_FACADE = ProcessorFacade(
            resources_provider=RESOURCES_PROVIDER)
        _LOG.debug('aws-syndicate has been initialized')
    except ClientError as e:
        raise InternalError(
            f'An unexpected error has occurred trying to init connection: {e}'
        )


def initialize_project_state(do_not_sync_state=False):
    from syndicate.core.project_state.sync_processor import sync_project_state
    global PROJECT_STATE
    if not PROJECT_STATE:
        if not ProjectState.check_if_project_state_exists(CONF_PATH):
            USER_LOG.warning(
                "Config is set and generated, but project state file does not "
                "exist."
            )
            USER_LOG.warning(
                "Generating project state file '.syndicate' from the existing "
                "structure..."
            )
            PROJECT_STATE = ProjectState.build_from_structure(CONFIG)
        else:
            PROJECT_STATE = ProjectState(project_path=CONFIG.project_path)

    if do_not_sync_state:
        return

    sync_project_state()


def validate_temp_credentials_expiration(expiration):
    if not isinstance(expiration, datetime):
        expiration = datetime.fromisoformat(expiration)
    expiration_datetime = expiration - timedelta(
        minutes=CREDENTIALS_EXPIRATION_THRESHOLD_MIN)
    now_datetime = datetime.now(timezone.utc)

    return expiration_datetime <= now_datetime


def prompt_mfa_code():
    mfa_code = input('Please enter your MFA code to generate '
                     'temp credentials: ')
    while 1:
        if len(mfa_code) == 6 and re.match('[0-9]{6}', mfa_code):
            break
        mfa_code = input(f'Token code must consist of 6 numbers. '
                         f'Try again: ')
    return mfa_code


def initialize_signal_handling():
    from signal import SIGINT, signal
    signal(SIGINT, handle_interruption)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import time

from boto3 import client

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry
from syndicate.core.helper import dict_keys_to_camel_case

_LOG = get_logger(__name__)

DATA_SOURCE_TYPE_CONFIG_MAPPING = {
    'AWS_LAMBDA': 'lambdaConfig',
    'AMAZON_DYNAMODB': 'dynamodbConfig',
    'AMAZON_ELASTICSEARCH': 'elasticsearchConfig',
    'HTTP': 'httpConfig',
    'RELATIONAL_DATABASE': 'relationalDatabaseConfig',
    'AMAZON_OPENSEARCH_SERVICE': 'openSearchServiceConfig',
    'AMAZON_EVENTBRIDGE': 'eventBridgeConfig'
}

REDUNDANT_RESOLVER_EXCEPTION_TEXT = 'Only one resolver is allowed per field'
DATA_SOURCE_EXISTS_EXCEPTION_TEXT = \
    'Data source with name {name} already exists'


@apply_methods_decorator(retry())
class AppSyncConnection(object):
    """ AWS AppSync connection class. """

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('appsync', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        self.region = region
        _LOG.debug('Opened new AppSync connection.')

# ------------------------ Create ------------------------

    def create_api(self, name, event_config, tags, owner):
        params = dict(
            name=name
        )
        if event_config:
            params['eventConfig'] = tags
        if tags:
            params['tags'] = tags
        if owner:
            params['ownerContact'] = owner

        return self.client.create_api(**params)

    def create_graphql_api(self, name: str, auth_type: str, tags: dict = None,
                           user_pool_config: dict = None,
                           open_id_config: dict = None,
                           lambda_auth_config: dict = None,
                           log_config: dict = None,
                           xray_enabled: bool = None,
                           extra_auth_types: list = None):
        params = dict(
            name=name,
            authenticationType=auth_type
        )
        if tags:
            params['tags'] = tags
        if user_pool_config:
            user_pool_config = dict_keys_to_camel_case(user_pool_config)
            params['userPoolConfig'] = user_pool_config
        if open_id_config:
            open_id_config = dict_keys_to_camel_case(open_id_config)
            params['openIDConnectConfig'] = open_id_config
        if lambda_auth_config:
            lambda_auth_config = dict_keys_to_camel_case(lambda_auth_config)
            params['lambdaAuthorizerConfig'] = lambda_auth_config
        if log_config:
            params['logConfig'] = dict_keys_to_camel_case(log_config)
        if xray_enabled:
            params['xrayEnabled'] = xray_enabled
        if extra_auth_types:
            params['additionalAuthenticationProviders'] = extra_auth_types

        return self.client.create_graphql_api(**params)['graphqlApi']['apiId']

    def create_schema(self, api_id: str, definition: str):
        response = self.client.start_schema_creation(
            apiId=api_id,
            definition=str.encode(definition)
        )
        status = response['status']
        details = response.get('details', '')
        while status == 'PROCESSING':
            time.sleep(2)
            response = self.client.get_schema_creation_status(
                apiId=api_id
            )
            status = response['status']
            details = response.get('details', '')
        return status, details

    def create_type(self, api_id: str, definition: str, format: str):
        params = dict(
            api_id=api_id,
            definition=definition,
            format=format
        )
        return self.client.create_type(**params)['type']

    def create_data_source(self, api_id: str, name: str, source_type: str,
                           source_config: dict = None, description: str = None,
                           service_role_arn: str = None):
        params = dict(
            apiId=api_id,
            name=name,
            type=source_type
        )
        config_key = DATA_SOURCE_TYPE_CONFIG_MAPPING.get(source_type)
        if config_key:
            source_config = dict_keys_to_camel_case(source_config)
            params[config_key] = source_config
        if description:
            params['description'] = description
        if service_role_arn:
            params['serviceRoleArn'] = service_role_arn

        try:
            return self.client.create_data_source(**params)['dataSource']
        except self.client.exceptions.BadRequestException as e:
            error_text = DATA_SOURCE_EXISTS_EXCEPTION_TEXT.format(name=name)
            if error_text in str(e):
                _LOG.warning(error_text)
                return
            else:
                raise e

    def create_function(self, api_id: str, func_params: dict):
        params = dict_keys_to_camel_case(func_params)
        params['apiId'] = api_id

        return self.client.create_function(**params)['functionConfiguration']

    def create_resolver(self, api_id: str, type_name: str, field_name: str,
                        kind: str, runtime: str = None,
                        data_source_name: str = None, code: str = None,
                        request_mapping_template: str = None,
                        response_mapping_template: str = None,
                        max_batch_size: int = None,
                        pipeline_config: dict = None):
        params = dict(
            apiId=api_id,
            typeName=type_name,
            fieldName=field_name,
            kind=kind
        )
        if runtime:
            params['runtime'] = runtime
        if data_source_name:
            params['dataSourceName'] = data_source_name
        if code:
            params['code'] = code
        if request_mapping_template:
            params['requestMappingTemplate'] = request_mapping_template
        if response_mapping_template:
            params['responseMappingTemplate'] = response_mapping_template
        if pipeline_config:
            params['pipelineConfig'] = pipeline_config
        if max_batch_size:
            params['maxBatchSize'] = max_batch_size

        try:
            return self.client.create_resolver(**params)['resolver']
        except self.client.exceptions.BadRequestException as e:
            if REDUNDANT_RESOLVER_EXCEPTION_TEXT in str(e):
                _LOG.warning(f'Only one resolver is allowed per field '
                             f'{field_name}; type {type_name}. '
                             f'Ignoring redundant resolver.')
                return
            else:
                raise e

    def create_api_key(self, api_id: str, description: str = None,
                       expires: int = None):
        params = dict(
            apiId=api_id
        )
        if description:
            params['description'] = description
        if expires:
            params['expires'] = expires

        return self.client.create_api_key(**params)['apiKey']

# ------------------------ Get ------------------------

    def get_graphql_api(self, api_id: str):
        return self.client.get_graphql_api(apiId=api_id)['graphqlApi']

    def get_data_source(self, api_id: str, name: str):
        try:
            return self.client.get_data_source(
                apiId=api_id, name=name)['dataSource']
        except self.client.exceptions.NotFoundException:
            _LOG.warning(f'Data source {name} not found')
            return

    def get_graphql_api_by_name(self, name):
        # TODO change list_graphql_apis to list_apis when upgrade boto3 version
        def process_apis(resume_token=None):
            pagination_conf = {
                'MaxItems': 60,
                'PageSize': 10
            }
            if resume_token:
                pagination_conf['StartingToken'] = resume_token
            response = paginator.paginate(
                PaginationConfig=pagination_conf
            )
            for page in response:
                apis.extend(
                    [api for api in page['graphqlApis'] if
                     api['name'] == name]
                )
            return response.resume_token

        apis = []
        paginator = self.client.get_paginator('list_graphql_apis')

        next_token = process_apis()
        while next_token:
            next_token = process_apis(next_token)

        if len(apis) == 1:
            return apis[0]
        if len(apis) > 1:
            _LOG.warn(f'AppSync API can\'t be identified unambiguously '
                      f'because there is more than one resource with the name '
                      f'"{name}" in the region {self.region}.')
        else:
            _LOG.warn(f'AppSync API with the name "{name}" '
                      f'not found in the region {self.region}')

    def get_resolver(self, api_id: str, type_name: str, field_name: str):
        try:
            return self.client.get_resolver(
                apiId=api_id,
                typeName=type_name,
                fieldName=field_name
            )['resolver']
        except self.client.exceptions.NotFoundException:
            _LOG.warning(f'Resolver for type {type_name} and field '
                         f'{field_name} not found')
            return

    def list_resolvers(self, api_id: str, type_name: str):
        result = []
        try:
            paginator = self.client.get_paginator('list_resolvers')
            for response in paginator.paginate(apiId=api_id,
                                               typeName=type_name):
                result.extend(response['resolvers'])
        except self.client.exceptions.NotFoundException:
            return

        return result

    def list_data_sources(self, api_id: str) -> list | None:
        result = []
        try:
            paginator = self.client.get_paginator('list_data_sources')
            for response in paginator.paginate(apiId=api_id):
                result.extend(response['dataSources'])
        except self.client.exceptions.NotFoundException:
            return

        return result

    def list_types(self, api_id: str) -> list | None:
        result = []
        try:
            paginator = self.client.get_paginator('list_types')
            for response in paginator.paginate(apiId=api_id, format='JSON'):
                result.extend(response['types'])
        except self.client.exceptions.NotFoundException:
            return

        return result

    def list_functions(self, api_id: str) -> list:
        result = []
        try:
            paginator = self.client.get_paginator('list_functions')
            for response in paginator.paginate(apiId=api_id):
                result.extend(response['functions'])
        except self.client.exceptions.NotFoundException:
            pass

        return result

    def get_schema(self, api_id: str, format: str = None):
        return self.client.get_introspection_schema(
            apiId=api_id,
            format='SDL' if not format else format
        )['schema']

    def list_api_keys(self, api_id: str) -> list:
        return self.client.list_api_keys(apiId=api_id)['apiKeys']

# ------------------------ Update ------------------------

    def update_data_source(self, api_id: str, name: str, source_type: str,
                           source_config: dict = None,
                           description: dict = None,
                           service_role_arn: str = None):
        params = dict(
            apiId=api_id,
            name=name,
            type=source_type
        )
        config_key = DATA_SOURCE_TYPE_CONFIG_MAPPING.get(source_type)
        if config_key:
            source_config = dict_keys_to_camel_case(source_config)
            params[config_key] = source_config
        if description:
            params['description'] = description
        if service_role_arn:
            params['serviceRoleArn'] = service_role_arn

        return self.client.update_data_source(**params)['dataSource']

    def update_function(self, api_id: str, function_id: str,
                        func_params: dict):
        params = dict_keys_to_camel_case(func_params)
        params.update({
            'apiId': api_id,
            'functionId': function_id
        })
        return self.client.update_function(**params)['functionConfiguration']

    def update_resolver(self, api_id: str, type_name: str, field_name: str,
                        kind: str, runtime: str = None,
                        data_source_name: str = None,
                        request_mapping_template: str = None,
                        response_mapping_template: str = None,
                        code: str = None, max_batch_size: int = None,
                        pipeline_config: dict = None):
        params = dict(
            apiId=api_id,
            typeName=type_name,
            fieldName=field_name,
            kind=kind
        )
        if runtime:
            params['runtime'] = runtime
        if data_source_name:
            params['dataSourceName'] = data_source_name
        if code:
            params['code'] = code
        if request_mapping_template:
            params['requestMappingTemplate'] = request_mapping_template
        if response_mapping_template:
            params['responseMappingTemplate'] = response_mapping_template
        if max_batch_size:
            params['maxBatchSize'] = max_batch_size
        if pipeline_config:
            params['pipelineConfig'] = pipeline_config

        return self.client.update_resolver(**params)['resolver']

    def update_graphql_api(self, api_id: str, name: str,
                           log_config: dict = None,
                           auth_type: str = None,
                           user_pool_config: dict = None,
                           open_id_config: dict = None,
                           lambda_auth_config: dict = None,
                           xray_enabled: bool = None,
                           extra_auth_types: list = None):
        params = dict(
            apiId=api_id,
            name=name
        )
        if auth_type:
            params['authenticationType'] = auth_type
        if user_pool_config:
            user_pool_config = dict_keys_to_camel_case(user_pool_config)
            params['userPoolConfig'] = user_pool_config
        if open_id_config:
            open_id_config = dict_keys_to_camel_case(open_id_config)
            params['openIDConnectConfig'] = open_id_config
        if lambda_auth_config:
            lambda_auth_config = dict_keys_to_camel_case(lambda_auth_config)
            params['lambdaAuthorizerConfig'] = lambda_auth_config
        if log_config:
            params['logConfig'] = dict_keys_to_camel_case(log_config)
        if xray_enabled:
            params['xrayEnabled'] = xray_enabled
        if extra_auth_types:
            params['additionalAuthenticationProviders'] = extra_auth_types

        return self.client.update_graphql_api(**params)['graphqlApi']['apiId']

# ------------------------ Delete ------------------------

    def delete_graphql_api(self, api_id: str):
        self.client.delete_graphql_api(apiId=api_id)

    def delete_data_source(self, api_id: str, name: str):
        return self.client.delete_data_source(
            apiId=api_id,
            name=name
        )

    def delete_function(self, api_id: str, func_id: str):
        try:
            return self.client.delete_function(
                apiId=api_id,
                functionId=func_id
            )
        except Exception as e:
            message = ('Cannot delete a function which is currently used by a '
                       'resolver')
            if message in str(e):
                _LOG.warn(str(e))
            else:
                raise

    def delete_resolver(self, api_id: str, type_name: str, field_name: str):

        return self.client.delete_resolver(
            apiId=api_id,
            typeName=type_name,
            fieldName=field_name
        )



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from time import time
from uuid import uuid1
from pathlib import PurePath
from botocore.exceptions import ClientError

from syndicate.exceptions import ResourceNotFoundError, ArtifactError
from syndicate.commons.log_helper import get_logger
from syndicate.core.build.meta_processor import S3_PATH_NAME
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger(__name__)


class EbsResource(BaseResource):

    def __init__(self, ec2_conn, iam_conn, ebs_conn, sns_conn,
                 s3_conn, region, account_id, deploy_target_bucket) -> None:
        self.ec2_conn = ec2_conn
        self.iam_conn = iam_conn
        self.ebs_conn = ebs_conn
        self.sns_conn = sns_conn
        self.s3_conn = s3_conn
        self.region = region
        self.account_id = account_id
        self.deploy_target_bucket = deploy_target_bucket

    def describe_ebs(self, name, meta, response=None):
        arn = f'arn:aws:elasticbeanstalk:{self.region}:{self.account_id}' \
            f':application/{name}'
        if not response:
            response = self.ebs_conn.describe_applications([name])
        if not response:
            return {}
        return {
            arn: build_description_obj(response, name, meta)
        }

    def create_ebs(self, args):
        return self.create_pool(self._create_ebs_app_env_from_meta, args)

    @unpack_kwargs
    def _create_ebs_app_env_from_meta(self, name, meta):
        from syndicate.core import CONFIG
        response = self.ebs_conn.describe_applications([name])
        if response:
            _LOG.warn(f'{name} EBS app exists.')
            return self.describe_ebs(name, meta, response[0])

        env_settings = meta['env_settings']
        topic_name = meta.get('notification_topic')
        # check topic exists
        if topic_name:
            topic_arn = self.sns_conn.get_topic_arn(topic_name)
            if topic_arn:
                env_settings.append({
                    "OptionName": "Notification Topic ARN",
                    "Namespace": "aws:elasticbeanstalk:sns:topics",
                    "Value": topic_arn
                })
            else:
                raise ResourceNotFoundError(
                    f"Cant find notification topic '{topic_name}' for EBS."
                )
        # check key pair exists
        key_pair_name = meta['ec2_key_pair']
        if self.ec2_conn.if_key_pair_exists(key_pair_name):
            env_settings.append({
                "OptionName": "KeyName",
                "ResourceName": "AWSEBAutoScalingLaunchConfiguration",
                "Namespace": "aws:cloudformation:template:resource:property",
                "Value": key_pair_name
            })
        else:
            raise ResourceNotFoundError(
                f"Specified key pair does not exist: '{key_pair_name}'."
            )
        # check ec2 role exists
        iam_role = meta['ec2_role']
        if self.iam_conn.check_if_role_exists(iam_role):
            env_settings.append({
                "OptionName": "IamInstanceProfile",
                "ResourceName": "AWSEBAutoScalingLaunchConfiguration",
                "Namespace": "aws:autoscaling:launchconfiguration",
                "Value": iam_role
            })
        else:
            raise ResourceNotFoundError(
                f"Specified iam role does not exist: '{iam_role}'."
            )
        # check service role exists
        iam_role = meta['ebs_service_role']
        if self.iam_conn.check_if_role_exists(iam_role):
            env_settings.append({
                "OptionName": "ServiceRole",
                "Namespace": "aws:elasticbeanstalk:environment",
                "Value": iam_role
            })
        else:
            raise ResourceNotFoundError(
                f"Specified iam role does not exist: '{iam_role}'."
            )
        image_id = meta.get('image_id')
        if image_id:
            env_settings.append({
                "OptionName": "ImageId",
                "ResourceName": "AWSEBAutoScalingLaunchConfiguration",
                "Namespace": "aws:autoscaling:launchconfiguration",
                "Value": image_id
            })
        else:
            _LOG.warn('Image id is not specified.')
        # check that desired solution stack exists
        stack = meta['stack']
        available_stacks = self.ebs_conn. \
            describe_available_solutions_stack_names()
        if stack not in available_stacks:
            raise ResourceNotFoundError(
                f'No solution stack named {stack} found.'
                f' Available:\n{available_stacks}'
            )
        vpc_id = next(
            (option for option in env_settings if
             option['OptionName'] == 'VPCId'),
            None)
        if not vpc_id:
            vpc_id = self.ec2_conn.get_default_vpc_id()
            _LOG.info('Default vpc id %s', vpc_id)
            if vpc_id:
                _LOG.debug('Will use vpc %s', vpc_id)
                subnets = self.ec2_conn.list_subnets(filters=[{
                    'Name': 'vpc-id',
                    'Values': [vpc_id]
                }])
                _LOG.debug(f'Found subnets for {vpc_id} vpc: {subnets}')
                if subnets:
                    _LOG.info(f'Will attach default {vpc_id} vpc to env')
                    self._add_subnets_info(env_settings, subnets, vpc_id)
                sg_id = self.ec2_conn.get_sg_id(group_name='default',
                                                vpc_id=vpc_id)
                if sg_id:
                    _LOG.debug(f'Found default sg with id {sg_id}')
                    env_settings.append({
                        "OptionName": "SecurityGroups",
                        "Namespace": "aws:autoscaling:launchconfiguration",
                        "Value": sg_id
                    })

        env_name = meta["env_name"] + str(int(time()))

        start = time()
        end = start + 180
        while end > time():
            describe_app_result = self.ebs_conn.describe_applications([name])
            if not describe_app_result:
                break

        # create APP
        response = self.ebs_conn.create_application(name,
                                                    tags=meta.get('tags'))
        _LOG.info(f'Created EBS app {name}.')
        # create ENV
        self.ebs_conn.create_environment(app_name=name,
                                         env_name=env_name,
                                         option_settings=env_settings,
                                         tier=meta['tier'],
                                         solution_stack_name=stack,
                                         tags=meta.get('tags'))
        key = meta[S3_PATH_NAME]
        key_compound = PurePath(CONFIG.deploy_target_bucket_key_compound,
                                key).as_posix()
        if not self.s3_conn.is_file_exists(self.deploy_target_bucket,
                                           key_compound):
            raise ArtifactError(
                f'Deployment package does not exist in '
                f'{self.deploy_target_bucket} bucket'
            )

        # create VERSION
        version_label = env_name + str(uuid1())
        self.ebs_conn.create_app_version(app_name=name,
                                         version_label=version_label,
                                         s3_bucket=self.deploy_target_bucket,
                                         s3_key=key_compound,
                                         tags=meta.get('tags'))
        _LOG.debug(f'Waiting for beanstalk env {env_name}')
        # wait for env creation
        start = time()
        status = {}
        end = start + 360  # end in 6 min
        while end > time():
            status = self.ebs_conn.describe_environment_health(
                env_name=env_name,
                attr_names=[
                    'Status'])
            if status['Status'] == 'Ready':
                _LOG.info('Launching env took %s.', time() - start)
                break
        if status['Status'] != 'Ready':
            _LOG.error(f'Env status: {status}. Failed to create env.')
        # deploy new app version
        self.ebs_conn.deploy_env_version(name, env_name, version_label)
        _LOG.info('Created environment for %s.', name)
        return self.describe_ebs(name, meta, response)

    @staticmethod
    def _add_subnets_info(env_settings, subnets, vpc_id):
        env_settings.append({
            "OptionName": "VPCId",
            "Namespace": "aws:ec2:vpc",
            "Value": vpc_id
        })
        subnets = ",".join(subnet['SubnetId'] for subnet in subnets)
        env_settings.append({
            "OptionName": "Subnets",
            "Namespace": "aws:ec2:vpc",
            "Value": subnets
        })

    def remove_ebs_apps(self, args):
        return self.create_pool(self._remove_ebs_app, args)

    @unpack_kwargs
    def _remove_ebs_app(self, arn, config):
        app_name = config['resource_name']
        try:
            self.ebs_conn.remove_app(app_name, log_not_found_error=False)
            _LOG.info(f'EBS app {app_name} was removed.')
            return {arn: config}
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                _LOG.warn(f'EBS app {app_name} is not found')
                return {arn: config}
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.exceptions import ResourceMetadataError
from syndicate.core.constants import IAM_AUTH_TYPE, NONE_AUTH_TYPE, \
    LAMBDA_ARCHITECTURE_LIST, DYNAMO_DB_TRIGGER, CLOUD_WATCH_RULE_TRIGGER, \
    DYNAMODB_TRIGGER_REQUIRED_PARAMS, EVENT_BRIDGE_RULE_TRIGGER, \
    CLOUD_WATCH_TRIGGER_REQUIRED_PARAMS, S3_TRIGGER, \
    S3_TRIGGER_REQUIRED_PARAMS, SNS_TOPIC_TRIGGER, SNS_TRIGGER_REQUIRED_PARAMS, \
    KINESIS_TRIGGER, KINESIS_TRIGGER_REQUIRED_PARAMS, SQS_TRIGGER, \
    SQS_TRIGGER_REQUIRED_PARAMS

RESOURCE_TYPE_KEY = 'resource_type'

TRIGGER_REQUIRED_PARAMS_MAPPING = {
        DYNAMO_DB_TRIGGER: DYNAMODB_TRIGGER_REQUIRED_PARAMS,
        CLOUD_WATCH_RULE_TRIGGER: CLOUD_WATCH_TRIGGER_REQUIRED_PARAMS,
        EVENT_BRIDGE_RULE_TRIGGER: CLOUD_WATCH_TRIGGER_REQUIRED_PARAMS,
        S3_TRIGGER: S3_TRIGGER_REQUIRED_PARAMS,
        SNS_TOPIC_TRIGGER: SNS_TRIGGER_REQUIRED_PARAMS,
        KINESIS_TRIGGER: KINESIS_TRIGGER_REQUIRED_PARAMS,
        SQS_TRIGGER: SQS_TRIGGER_REQUIRED_PARAMS
    }


class LambdaValidator:
    def __init__(self, name, meta):
        self._name = name
        self._meta = meta

    def validate(self):
        url_config = self._meta.get('url_config')
        if url_config:
            self._validate_url_config(url_config)

        ephemeral_storage = self._meta.get('ephemeral_storage')
        if ephemeral_storage:
            self._validate_ephemeral_storage(ephemeral_storage)

        architectures = self._meta.get('architectures')
        if architectures:
            self._validate_architecture(architectures)

        event_sources = self._meta.get('event_sources')
        if event_sources:
            self._validate_event_sources(event_sources)

    def _error(self, message):
        raise ResourceMetadataError(message)

    def _validate_url_config(self, url_config):
        if 'auth_type' not in url_config:
            self._error('\'auth_type\' is required in lambda\'s url config')
        auth_types = [IAM_AUTH_TYPE, NONE_AUTH_TYPE]
        if url_config['auth_type'] not in auth_types:
            self._error('\'auth_type\' must be equal to one of these: '
                        f'{", ".join(auth_types)}')
        cors = url_config.get('cors')
        if cors and not isinstance(cors, dict):
            self._error('\'cors\' parameter in lambda config must be a dict')
        if cors:
            allowed_parameters = {'allow_credentials',
                                  'allow_headers', 'allow_methods',
                                  'allow_origins', 'expose_headers',
                                  'max_age'}
            impostors = set(cors.keys()) - allowed_parameters
            if impostors:
                self._error(f'Only these parameters are allowed: '
                            f'{allowed_parameters}')

    def _validate_ephemeral_storage(self, ephemeral_storage):
        if not isinstance(ephemeral_storage, int):
            self._error(f'Ephemeral storage size must an integer but not '
                        f'\'{type(ephemeral_storage).__name__}\'')
        if not (512 <= ephemeral_storage <= 10240):
            self._error('Ephemeral storage size must be between '
                        '512 and 10240 MB')

    def _validate_architecture(self, architectures):
        for architecture in architectures:
            if architecture not in LAMBDA_ARCHITECTURE_LIST:
                self._error(f'Specified unsupported architecture: '
                            f'"{architecture}". Currently supported '
                            f'architectures: {LAMBDA_ARCHITECTURE_LIST}')

    def _validate_event_sources(self, event_sources: list[dict]) -> None:
        errors = []
        for index, event_source in enumerate(event_sources, 1):
            if not isinstance(event_source, dict):
                errors.append(
                    f"Event source configuration with index '{index}' is "
                    f"invalid. Expected type: 'map(object)' actual: "
                    f"'{type(event_source).__name__}'"
                )
                continue

            resource_type = event_source.get(RESOURCE_TYPE_KEY)
            if resource_type is None:
                errors.append(
                    f"Event source configuration with index '{index}' is "
                    f"invalid. The parameter '{RESOURCE_TYPE_KEY}' is "
                    f"mandatory."
                )
                continue
            if resource_type not in TRIGGER_REQUIRED_PARAMS_MAPPING:
                errors.append(
                    f"Event source configuration with index '{index}' is "
                    f"invalid. Trigger type '{resource_type}' is not "
                    f"supported. Supported trigger types: "
                    f"{list(TRIGGER_REQUIRED_PARAMS_MAPPING.keys())}"
                )
                continue

            req_params = TRIGGER_REQUIRED_PARAMS_MAPPING[resource_type]
            req_params.insert(0, RESOURCE_TYPE_KEY)
            existing_params = list(event_source.keys())
            for each in req_params:
                if each not in existing_params:
                    errors.append(
                        f"Event source configuration with index '{index}' is "
                        f"invalid. Not all required parameters specified. "
                        f"Required parameters: "
                        f"{req_params}. "
                        f"Given parameters: {existing_params}."
                    )
                    break

        if errors:
            errors_string = '\n'.join(errors)
            self._error(
                f"Lambda '{self._name}' event sources haven't passed "
                f"validation.\n{errors_string}")


def validate_lambda(name, meta):
    LambdaValidator(name, meta).validate()



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_resource_reference_builder import \
    build_role_arn_ref


class BatchJobDefConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        job_definition_type = resource.get('job_definition_type')
        container_properties = resource.get('container_properties')
        parameters = resource.get('parameters')
        timeout = resource.get('timeout')
        tags = resource.get('tags')
        retry_strategy = resource.get('retry_strategy')

        batch_job_definition = batch_job_def(job_def_name=name,
                                             job_def_type=job_definition_type,
                                             container_properties=container_properties,
                                             parameters=parameters,
                                             timeout=timeout, tags=tags,
                                             retry_strategy=retry_strategy)
        self.template.add_aws_batch_job_definition(meta=batch_job_definition)


def batch_job_def(job_def_name, job_def_type, container_properties,
                  parameters, tags, timeout, retry_strategy):
    params = {
        'name': job_def_name,
        'type': job_def_type
    }

    if parameters:
        params['parameters'] = parameters
    if tags:
        params['tags'] = tags
    if timeout:
        params['timeout'] = timeout
    if retry_strategy:
        params['retry_strategy'] = retry_strategy

    if container_properties:
        properties = {}

        image = container_properties.get('image')
        if image:
            properties['image'] = image
        vcpus = container_properties.get('vcpus')
        if vcpus:
            properties['vcpus'] = vcpus
        memory = container_properties.get('memory')
        if memory:
            properties['memory'] = memory
        command = container_properties.get('command', [])
        if command:
            properties['command'] = command
        job_role_arn = container_properties.get('job_role_arn')
        if job_role_arn:
            role_arn_ref = build_role_arn_ref(role_name=job_role_arn)
            properties['executionRoleArn'] = role_arn_ref

        if properties:
            params['container_properties'] = json.dumps(properties)

    resource = {
        job_def_name: params
    }
    return resource



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import ipaddress
import re
import string
from typing import Optional

from syndicate.exceptions import InvalidValueError, InvalidTypeError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core import ClientError
from syndicate.core.constants import S3_BUCKET_ACL_LIST
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def validate_bucket_name(bucket_name: str):
    """Checks whether the given bucket name is valid.
    If the given name isn't valid, ValueError with an appropriate message
    is raised.
    :type bucket_name: str
    :param bucket_name: the name to check
    """
    bucket_name = bucket_name.strip()
    _LOG.info(f"Starting validating bucket name '{bucket_name}'")
    error = None

    raw_bucket_name, _ = (bucket_name.split('/', 1)
                          if bucket_name and '/' in bucket_name
                          else (bucket_name, None))
    if not raw_bucket_name or not 3 <= len(raw_bucket_name) <= 63:
        error = 'Bucket name must be between 3 and 63 characters long'
    else:
        invalid_characters = re.search('[^a-z0-9.-]', bucket_name)
        if invalid_characters:
            character = invalid_characters.group()
            if character in string.ascii_uppercase:
                error = 'Bucket name must not contain uppercase characters'
            else:
                error = f'Bucket name contains invalid characters: {character}'
        elif any(bucket_name.startswith(symbol) for symbol in '.-'):
            error = 'Bucket name must start with a lowercase letter or number'
        elif any(bucket_name.endswith(symbol) for symbol in '.-'):
            error = 'Bucket name must not end with dash or period'
        elif '..' in bucket_name:
            error = 'Bucket name must not contain two adjacent periods'
        elif '.-' in bucket_name or '-.' in bucket_name:
            error = 'Bucket name must not contain dash next to period'
        else:
            try:
                ipaddress.ip_address(bucket_name)
                error = 'Bucket name must not resemble an IP address'
            except ValueError:
                pass
    if error:
        _LOG.warning(error)
        raise InvalidValueError(error)
    _LOG.info(f"Finished validating bucket name '{bucket_name}'")


class S3Resource(BaseResource):

    def __init__(self, s3_conn, account_id) -> None:
        self.s3_conn = s3_conn
        self.account_id = account_id

    def create_s3_bucket(self, args):
        return self.create_pool(self._create_s3_bucket_from_meta, args)

    def describe_bucket(self, name, meta):
        arn = self.get_bucket_arn(name)
        acl_response = self.s3_conn.get_bucket_acl(name)
        location_response = self.s3_conn.get_bucket_location(name)
        bucket_policy = self.s3_conn.get_bucket_policy(name)
        if not location_response:
            return {}
        response = {
            'bucket_acl': acl_response,
            'location': location_response,
        }
        if bucket_policy:
            response['policy'] = bucket_policy
        return {
            arn: build_description_obj(response, name, meta)
        }

    @staticmethod
    def get_bucket_arn(name):
        return 'arn:aws:s3:::{0}'.format(name)

    @unpack_kwargs
    def _create_s3_bucket_from_meta(self, name, meta):
        if self.s3_conn.is_bucket_exists(name):
            _LOG.warn('{0} bucket exists.'.format(name))
            return self.describe_bucket(name, meta)

        self.s3_conn.create_bucket(name, location=meta.get('location'))
        _LOG.info('Created S3 bucket {0}.'.format(name))

        public_access_block = meta.get('public_access_block', {})
        if not all([isinstance(param, bool) for param in
                    public_access_block.values()]):
            message = f'Parameters inside public_access_block should have ' \
                      f'bool type'
            _LOG.error(message)
            raise InvalidTypeError(message)
        self.s3_conn.put_public_access_block(name,
                                             **public_access_block)

        acl = meta.get('acl')
        if acl:
            if acl not in S3_BUCKET_ACL_LIST:
                raise InvalidValueError(
                    f"Invalid value of S3 bucket ACL! Must be one of the "
                    f"'{S3_BUCKET_ACL_LIST}'"
                )
            self.s3_conn.put_bucket_acl(name, acl)

        policy = meta.get('policy')
        if policy:
            self.s3_conn.add_bucket_policy(name, policy)
            _LOG.debug('Policy on {0} S3 bucket is set up.'.format(name))

        website_hosting = meta['website_hosting'].get('enabled') \
            if meta.get('website_hosting') else None
        if website_hosting:
            index_document = meta['website_hosting'].get('index_document')
            error_document = meta['website_hosting'].get('error_document')
            if not all([isinstance(param, str) for param in (index_document,
                                                             error_document)]):
                raise InvalidTypeError(
                    "Value of parameters 'index_document' and "
                    "'error_document' must be of 'str' type"
                )
            self.s3_conn.enable_website_hosting(name,
                                                index_document,
                                                error_document)
            _LOG.debug(f'Website hosting configured with parameters: '
                       f'\'index_document\': \'{index_document}\', '
                       f'\'error_document\': \'{error_document}\'')
            website_endpoint = (
                f'http://{name}.s3-website.{self.s3_conn.region}.amazonaws.com'
                f'/{index_document}')
            USER_LOG.info(f'Bucket website endpoint: {website_endpoint}')

        rules = meta.get('LifecycleConfiguration')
        if rules:
            self.s3_conn.add_bucket_rule(name, rules)
            _LOG.debug('Rules on {0} S3 bucket are set up.'.format(name))

        cors_configuration = meta.get('cors')
        if cors_configuration:
            self.s3_conn.put_cors(bucket_name=name, rules=cors_configuration)
        return self.describe_bucket(name, meta)

    def _delete_objects(self, bucket_name, keys):
        response = self.s3_conn.delete_objects(bucket_name, keys)
        errors = response.get('Errors')
        if errors:
            error_keys = [{
                'Key': i['Key'],
                'VersionId': i['VersionId']
            } for i in errors]
            return error_keys
        else:
            return []

    def remove_buckets(self, args):
        return self.create_pool(self._remove_bucket, args)

    @unpack_kwargs
    def _remove_bucket(self, arn, config):
        bucket_name = config['resource_name']
        try:
            self.s3_conn.remove_bucket(bucket_name=bucket_name,
                                       log_not_found_error=False)
            _LOG.info('S3 bucket {0} was removed.'.format(bucket_name))
            return {arn: config}
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchBucket':
                _LOG.warn('S3 bucket {0} is not found'.format(bucket_name))
                return {arn: config}
            else:
                raise e

    def build_bucket_arn(self, maybe_arn: str) -> Optional[str]:
        if not isinstance(maybe_arn, str):
            return
        if self.is_bucket_arn(maybe_arn):
            return maybe_arn
        return f'arn:aws:s3:::{maybe_arn}'

    @staticmethod
    def is_bucket_arn(maybe_arn: str) -> bool:
        arn_regex = r'^arn:aws:s3:::[a-z0-9.-]{3,63}(?:/.*)?$'
        return bool(re.match(arn_regex, maybe_arn))



from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import RDS_DB_CLUSTER_TYPE, RDS_DB_INSTANCE_TYPE
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.exceptions import InvalidValueError, ParameterError

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


class RDSDBClusterGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = RDS_DB_CLUSTER_TYPE
    CONFIGURATION = {
        'engine': str,
        'engine_version': None,
        'master_username': str,
        'master_user_password': None,
        'database_name': str,
        'port': int,
        'iam_db_auth': None,
        'vpc_security_group_ids': list,
        'availability_zones': None,
        'db_subnet_group_name': None,
        'manage_master_user_password': None,
        'tags': dict
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _generate_resource_configuration(self) -> dict:

        if not self._dict.get('port'):
            if self._dict['engine'] == 'aurora-postgresql':
                self._dict['port'] = 5432
            else:
                self._dict['port'] = 3306

        self._dict['master_user_password'] = self._dict.pop('master_password')
        self._dict['manage_master_user_password'] = \
            self._dict.pop('manage_master_password', None)

        self._dict['db_subnet_group_name'] = \
            self._dict.pop('db_subnet_group', None)

        if not any((self._dict.get('master_user_password'),
                    self._dict.get('manage_master_user_password'))):
            raise ParameterError(
                "Either 'master_password' or 'manage_master_password' as "
                "`true` must be specified"
            )

        if self._dict.get('master_user_password'):
            self._validate_master_password()

        return super()._generate_resource_configuration()

    def _validate_master_password(self):
        to_validate = self._dict.get('master_user_password')
        _LOG.info(f"Validating master password '{to_validate}'...")
        error = None
        if not 8 <= len(to_validate) <= 100:
            error = "master password must contain from 8 to 100 characters"
        elif any(char in to_validate for char in ('"@/')):
            error = 'master password cannot contain forward slash (/), ' \
                    'double quote (") or the "at" symbol (@)'
        if error:
            raise InvalidValueError(error)
        _LOG.info("Master password validation passed successfully")


class RDSDBInstanceGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = RDS_DB_INSTANCE_TYPE
    CONFIGURATION = {
        'instance_class': str,
        'engine': str,
        'engine_version': None,
        'cluster_name': None,
        'database_name': None,
        'master_username': None,
        'master_user_password': None,
        'port': None,
        'publicly_accessible': False,
        'vpc_security_group_ids': None,
        'availability_zone': None,
        'tags': dict
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _generate_resource_configuration(self) -> dict:

        if self._dict.get('master_password'):
            self._validate_master_password()
            self._dict['master_user_password'] = \
                self._dict.pop('master_password')

        if cluster_name := self._dict.get('cluster_name'):
            self._validate_resource_existence(
                resource_name=cluster_name,
                resource_type=RDS_DB_CLUSTER_TYPE)

            deployment_resources = self._get_deployment_resources_file_content(
                self._get_resource_meta_paths(
                    resource_name=cluster_name,
                    resource_type=RDS_DB_CLUSTER_TYPE)[0])

            cluster_config = deployment_resources[cluster_name]
            self._dict['engine'] = cluster_config.get('engine')

        return super()._generate_resource_configuration()

    def _validate_master_password(self):
        to_validate = self._dict.get('master_password')
        _LOG.info(f"Validating master password '{to_validate}'...")
        error = None
        if not 8 <= len(to_validate) <= 100:
            error = "master password must contain from 8 to 100 characters"
        elif any(char in to_validate for char in ('"@/')):
            error = 'master password cannot contain forward slash (/), ' \
                    'double quote (") or the "at" symbol (@)'
        if error:
            raise InvalidValueError(error)
        _LOG.info("Master password validation passed successfully")


"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent.futures
import getpass
import hashlib
import json
import os
import re
import subprocess
import sys
import logging
import traceback
import zipfile
from datetime import datetime, timedelta
from functools import wraps
from pathlib import Path
from signal import SIGINT
from threading import Thread
from time import time
from typing import Union

import click
from click import BadParameter, MissingParameter
from tqdm import tqdm

from syndicate.exceptions import ArtifactAssemblingError, \
    InternalError, ProjectStateError, InvalidValueError, \
    SyndicateBaseError
from syndicate.commons.log_helper import get_logger, get_user_logger, \
    LOG_NAME, USER_LOG_NAME
from syndicate.core.conf.processor import path_resolver
from syndicate.core.conf.validator import ConfigValidator, ALL_REGIONS
from syndicate.core.constants import (BUILD_META_FILE_NAME,
                                      DEFAULT_SEP, DATE_FORMAT_ISO_8601,
                                      CUSTOM_AUTHORIZER_KEY, OK_RETURN_CODE,
                                      ABORTED_RETURN_CODE, FAILED_RETURN_CODE,
                                      PROFILER_ACTION, UPDATE_ACTION,
                                      WARMUP_ACTION)
from syndicate.core.project_state.project_state import MODIFICATION_LOCK, \
    WARMUP_LOCK, ProjectState
from syndicate.core.project_state.sync_processor import sync_project_state

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()

CONF_PATH = os.environ.get('SDCT_CONF')


def unpack_kwargs(handler_func):
    """ Decorator for unpack kwargs.

    :type handler_func: func
    :param handler_func: function which will be decorated
    """

    @wraps(handler_func)
    def wrapper(*kwargs):
        """ Wrapper func."""
        parameters = {}
        for i in kwargs:
            if type(i) is dict:
                parameters = i
        return handler_func(**parameters)

    return wrapper


def failed_status_code_on_exception(handler_func):
    """ Decorator to catch all exceptions and fail stage execution

    :type handler_func: func
    :param handler_func: function which will be decorated
    """

    @wraps(handler_func)
    def wrapper(*args, **kwargs):
        """ Wrapper func."""
        try:
            return handler_func(*args, **kwargs)
        except Exception as e:
            if isinstance(e, BadParameter):
                message = f"{e.__class__.__name__} {e.message}"
            elif isinstance(e, SyndicateBaseError):
                message = f"{e.__class__.__name__} occurred: {str(e)}"
            else:
                message = (f'An unexpected error occurred: '
                           f'{e.__class__.__name__} {str(e)}')

            USER_LOG.error(message)
            _LOG.exception(traceback.format_exc())

            return FAILED_RETURN_CODE

    return wrapper


def prettify_json(obj):
    return json.dumps(obj, indent=4)


def execute_command_by_path(command, path, shell=True):
    result = subprocess.run(command, shell=shell, cwd=path,
                            capture_output=True, text=True)

    pretty_command = \
        ' '.join(command) if isinstance(command, list) else command
    if result.returncode != 0:
        msg = (f'While running the command "{pretty_command}" occurred an '
               f'error:\n"{result.stdout}\n{result.stderr}"')
        raise ArtifactAssemblingError(msg)
    _LOG.info(f'Running the command "{pretty_command}"\n{result.stdout}'
              f'\n{result.stderr}')


def build_path(*paths):
    return DEFAULT_SEP.join(paths)


def _find_alias_and_replace(some_string):
    """ Find placeholder for alias in string. If found - replace with alias
    value.

    :type some_string: str
    """
    from syndicate.core import CONFIG
    first_index = some_string.index('${')
    second_index = some_string.index('}')
    alias_name = some_string[first_index + 2:second_index]
    res_alias = CONFIG.resolve_alias(alias_name)
    if not res_alias:
        raise InvalidValueError(f"Can not find alias for '{alias_name}'")
    result = (
            some_string[:first_index] + res_alias + some_string[
                                                    second_index + 1:])
    return result


def resolve_aliases_for_string(string_value):
    """ Look for aliases in string.

    :type string_value: str
    """
    input_string = string_value
    try:
        if '${' in string_value:
            if string_value.count('${') == string_value.count('}'):
                while True:
                    input_string = _find_alias_and_replace(input_string)
            else:
                raise InvalidValueError(
                    f"Broken alias in value: '{string_value}'."
                )
        return input_string
    except ValueError:
        return input_string


def check_required_param(ctx, param, value):
    if not value:
        raise BadParameter('Parameter is required')
    return value


def param_to_lower(ctx, param, value):
    if isinstance(value, tuple):
        return tuple([a_value.lower() for a_value in value])
    if isinstance(value, str):
        return value.lower()


def resolve_path_callback(ctx, param, value):
    if not value:
        raise BadParameter('Parameter is required')
    return path_resolver(value)


def generate_default_bundle_name(ctx, param, value):
    if value:
        return value
    from syndicate.core import CONFIG
    if CONFIG is None:
        USER_LOG.error('Configuration is not initialized. '
                       'Please check your configuration.')
        sys.exit(ABORTED_RETURN_CODE)

    # regex to replace all special characters except dash, underscore and dot
    pattern = re.compile('[^0-9a-zA-Z.\-_]')
    project_path = CONFIG.project_path
    project_name = project_path.split("/")[-1]
    result_project_name = re.sub(pattern, '', project_name)
    date = datetime.now().strftime("%y%m%d.%H%M%S")
    bundle_name = f'{result_project_name}_{date}'
    if len(bundle_name) > 63:
        USER_LOG.warn(f'Bundle name \'{bundle_name}\' is too long. Trim it to '
                      f'the last 63 characters. Please rename project to '
                      f'shorter name to avoid this warning.')
        return bundle_name[-63:]
    return bundle_name


def resolve_default_bundle_name(command_name):
    from syndicate.core import PROJECT_STATE
    if command_name in (PROFILER_ACTION, WARMUP_ACTION):
        bundle_name = PROJECT_STATE.latest_deployed_bundle_name
    else:
        bundle_name = PROJECT_STATE.latest_bundle_name
    if not bundle_name:
        USER_LOG.error(
            'Property \'bundle\' could not be resolved from the syndicate '
            'project state file.'
        )
        return ABORTED_RETURN_CODE
    return bundle_name


def resolve_default_deploy_name(command_name):
    from syndicate.core import PROJECT_STATE
    if command_name in (PROFILER_ACTION, UPDATE_ACTION, WARMUP_ACTION):
        deploy_name = PROJECT_STATE.latest_deployed_deploy_name
    else:
        deploy_name = PROJECT_STATE.default_deploy_name

    return deploy_name


param_resolver_map = {
    'bundle_name': resolve_default_bundle_name,
    'deploy_name': resolve_default_deploy_name
}


def resolve_default_value(ctx, param, value):
    if value:
        return value
    sync_project_state()
    command_name = ctx.info_name
    param_resolver = param_resolver_map.get(param.name)
    if not param_resolver:
        raise InternalError(
            f"There is no resolver of default value "
            f"for param {param.human_readable_name}")
    resolved_value = param_resolver(command_name=command_name)
    USER_LOG.info(
        f'Resolved value of {param.human_readable_name}: {resolved_value}')
    return resolved_value


def create_bundle_callback(ctx, param, value):
    from syndicate.core.build.helper import resolve_bundle_directory
    if not value:
        raise BadParameter('Parameter is required')
    bundle_path = resolve_bundle_directory(value)
    if not os.path.exists(bundle_path):
        os.makedirs(bundle_path)
    return value


def verify_bundle_callback(ctx, param, value):
    from syndicate.core.build.helper import resolve_bundle_directory
    bundle_path = resolve_bundle_directory(value)
    if not os.path.exists(bundle_path):
        raise click.BadParameter(
            "Bundle name does not exist. Please, invoke "
            "'syndicate assemble' command to create a bundle.")
    return value


def verify_meta_bundle_callback(ctx, param, value):
    from syndicate.core.build.helper import resolve_bundle_directory
    bundle_path = resolve_bundle_directory(value)
    build_meta_path = os.path.join(bundle_path, BUILD_META_FILE_NAME)
    if not os.path.exists(build_meta_path):
        raise click.BadParameter(
            "Bundle name is incorrect. {0} does not exist. Please, invoke "
            "'package_meta' command to create a file.".format(
                BUILD_META_FILE_NAME))
    return value


def resolve_and_verify_bundle_callback(ctx, param, value):
    if not value:
        _LOG.debug(
            f'{param.human_readable_name} is not specified, latest build will '
            f'be used')
        value = resolve_default_value(ctx, param, value)
        if not value:
            raise click.BadParameter(
                f'Couldn\'t resolve the parameter automatically. '
                f'Try to specify it manually'
            )
    return verify_meta_bundle_callback(ctx, param, value)


def write_content_to_file(file_path, file_name, obj):
    file_name = os.path.join(file_path, file_name)
    if os.path.exists(file_name):
        _LOG.warn('{0} already exists'.format(file_name))
    else:
        folder_path = Path(file_path)
        folder_path.mkdir(parents=True, exist_ok=True)
        meta_file = Path(file_name)
        meta_file.touch(exist_ok=True)
        with open(file_name, 'w+') as meta_file:
            json.dump(obj, meta_file)
        _LOG.info('{0} file was created.'.format(meta_file.name))


def sync_lock(lock_type):
    def real_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            sync_project_state()
            from syndicate.core import PROJECT_STATE
            if PROJECT_STATE.is_lock_free(lock_type):
                PROJECT_STATE.acquire_lock(lock_type)
                sync_project_state()
            else:
                raise ProjectStateError(
                    f"The project '{lock_type}' is locked. Run the command "
                    f"'syndicate status' for more details."
                )
            try:
                result = func(*args, **kwargs)
            except Exception as e:
                _LOG.exception("Error occurred: %s", str(e))
                from syndicate.core import PROJECT_STATE
                PROJECT_STATE.release_lock(lock_type)
                sync_project_state()
                raise
            PROJECT_STATE.release_lock(lock_type)
            sync_project_state()
            return result

        return wrapper

    return real_wrapper


def timeit(action_name=None):
    def internal_timeit(func):
        @wraps(func)
        def timed(*args, **kwargs):
            ts = time()
            result = func(*args, **kwargs)
            te = time()
            _LOG.info('Stage %s, elapsed time: %s', func.__name__,
                      str(timedelta(seconds=te - ts)))
            result_action_name = result.get('operation') if \
                isinstance(result, dict) else None
            if result_action_name:
                result = result.get('return_code', OK_RETURN_CODE)
            if action_name:
                username = getpass.getuser()
                duration = round(te - ts, 3)
                start_date_formatted = datetime.fromtimestamp(ts) \
                    .strftime(DATE_FORMAT_ISO_8601)
                end_date_formatted = datetime.fromtimestamp(te) \
                    .strftime(DATE_FORMAT_ISO_8601)

                bundle_name = kwargs.get('bundle_name')
                deploy_name = kwargs.get('deploy_name')
                rollback_on_error = kwargs.get('rollback_on_error')
                from syndicate.core import PROJECT_STATE
                PROJECT_STATE.log_execution_event(
                    operation=result_action_name or action_name,
                    initiator=username,
                    bundle_name=bundle_name,
                    deploy_name=deploy_name,
                    time_start=start_date_formatted,
                    time_end=end_date_formatted,
                    duration_sec=duration,
                    status=result,
                    rollback_on_error=rollback_on_error
                )
            return result

        return timed

    return internal_timeit


def execute_parallel_tasks(*fns):
    threads = []
    for fn in fns:
        t = Thread(target=fn)
        t.start()
        threads.append(t)
    for t in threads:
        t.join()


def handle_futures_progress_bar(futures):
    kwargs = {
        'total': len(futures),
        'unit': 'nap',
        'leave': True
    }
    for _ in tqdm(concurrent.futures.as_completed(futures), **kwargs):
        pass


def string_to_camel_case(s: str):
    temp = s.split('_')
    res = temp[0] + ''.join(ele.title() for ele in temp[1:])
    return str(res)


def string_to_upper_camel_case(s: str):
    temp = s.split('_')
    res = ''.join(ele.title() for ele in temp)
    return str(res)


def dict_keys_to_camel_case(d: dict):
    return _inner_dict_keys_to_camel_case(d, string_to_camel_case)


def dict_keys_to_upper_camel_case(d: dict):
    return _inner_dict_keys_to_camel_case(d, string_to_upper_camel_case)


def _inner_dict_keys_to_camel_case(d: dict, case_formatter):
    new_d = {}
    for key, value in d.items():
        if isinstance(value, (str, int, float)):
            new_d[case_formatter(key)] = value

        if isinstance(value, list):
            new_list = []
            for index, item in enumerate(value):
                if isinstance(item, (str, int, float)):
                    new_list.append(item)
                if isinstance(item, dict):
                    new_list.append(dict_keys_to_camel_case(item))
            new_d[case_formatter(key)] = new_list

        if isinstance(value, dict):
            new_d[case_formatter(key)] = dict_keys_to_camel_case(value)

    return new_d


def string_to_capitalized_camel_case(s: str):
    temp = s.split('_')
    res = ''.join(ele.capitalize() for ele in temp)
    return res


def dict_keys_to_capitalized_camel_case(d: dict):
    new_d = {}
    for key, value in d.items():
        if isinstance(value, (str, int)):
            new_d[string_to_capitalized_camel_case(key)] = value

        if isinstance(value, list):
            new_list = []
            for index, item in enumerate(value):
                if isinstance(item, (str, int)):
                    new_list.append(item)
                if isinstance(item, dict):
                    new_list.append(dict_keys_to_capitalized_camel_case(item))
            new_d[string_to_capitalized_camel_case(key)] = new_list

        if isinstance(value, dict):
            new_d[string_to_capitalized_camel_case(key)] = \
                dict_keys_to_capitalized_camel_case(value)

    return new_d


class OptionRequiredIf(click.Option):
    def __init__(self, *args, **kwargs):
        self.required_if = kwargs.pop('required_if')
        self.required_if_values = kwargs.pop('required_if_values', [])
        if not self.required_if:
            raise InternalError("'required_if' param must be specified")
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        is_current_present: bool = self.name in opts
        if self.required_if_values:
            is_required_ok: bool = \
                opts.get(self.required_if.replace('-','_')) in self.required_if_values
            message = (
                f"Option '{self.human_readable_name}' and '{self.required_if}' "
                f"must be specified together, and '{self.required_if}' must "
                f"have one of the next values {self.required_if_values}")
        else:
            is_required_ok: bool = self.required_if in opts
            message = (f"options: '{self.human_readable_name}' and "
                       f"'{self.required_if}' must be specified together")
        if is_current_present ^ is_required_ok:
            raise click.UsageError(message)
        else:
            return super().handle_parse_result(ctx, opts, args)


class MultiWordOption(click.Option):
    def __init__(self, *args, **kwargs):
        # Find name with "-" and add alias with "_"
        new_args = []

        for opt in args:
            alias_opts = []
            for alias in opt:
                if alias.startswith('--'):
                    underscored = alias[:2] + alias[2:].replace('-', '_')
                elif alias.startswith('-'):
                    underscored = alias[:1] + alias[1:].replace('-', '_')
                else:
                    underscored = alias.replace('-', '_')

                if underscored not in opt:
                    alias_opts.append(underscored)
            new_args.append(tuple(list(opt) + alias_opts))

        super().__init__(*new_args, **kwargs)

    def get_help_record(self, ctx):
        """
        Overrides the display of options in help, hiding aliases with
        underscores.
        """
        help_record = super().get_help_record(ctx)
        if help_record is None:
            return

        option_names = self.opts

        filtered = list(set([
            name.replace('_', '-') for name in option_names
        ]))

        if not self.is_flag and not self.count:
            filtered[-1] += f' {self.make_metavar()}'

        return ', '.join(filtered), help_record[1]

    @property
    def human_readable_name(self):
        """
        Overrides the human_readable_name to show the original name with dashes.
        """
        return self.name.replace('_', '-')


def combine_option_classes(*classes):
    """
    Combine multiple Click option classes into one.
    The order of classes matters, methods overlap in order of occurrence.
    """

    class CombinedOption(*classes, click.Option):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    return CombinedOption


class AliasedCommandsGroup(click.Group):
    """
    Custom Click Group to support command aliases.
    """

    def parse_args(self, ctx, args):
        current_cmd = self
        cmd_chain = []
        while args:
            next_arg = args[0]
            cmd = current_cmd.get_command(ctx, next_arg)
            if cmd is None:
                break
            cmd_chain.append(next_arg)
            args.pop(0)
            current_cmd = cmd
            if not isinstance(cmd, click.Group):
                break

        if isinstance(current_cmd, click.Group) and not args:
            sys.argv.append('--help')

        return super().parse_args(ctx, cmd_chain + args)

    def add_command(self, cmd, name=None):
        name = name or cmd.name
        super().add_command(cmd, name)

        alias_name = name.replace('-', '_')
        if alias_name != name and alias_name not in self.commands:
            self.commands[alias_name] = cmd

    def format_commands(self, ctx, formatter):
        """Format command list to hide aliases with underscore"""
        rows = []
        for subcommand in self.list_commands(ctx):
            command = self.get_command(ctx, subcommand)
            if command is None:
                continue
            help_text = command.get_short_help_str()

            if '-' in subcommand or '_' not in subcommand:
                rows.append((subcommand, help_text))
        if rows:
            with formatter.section('Commands'):
                formatter.write_dl(rows)

    def get_command(self, ctx, cmd_name):
        # get aliases
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        matches = [
            name for name, cmd in self.commands.items()
            if hasattr(cmd, 'aliases') and cmd_name in cmd.aliases
        ]
        if not matches:
            return None
        return self.commands[matches[0]]


class ValidRegionParamType(click.types.StringParamType):
    ALL_VALUE = 'ALL'
    name = 'region'

    def __init__(self, allowed_all=False):
        self.allowed_all = allowed_all

    def convert(self, value, param, ctx):
        value = super().convert(value, param, ctx)
        if self.allowed_all and value.upper() == self.ALL_VALUE:
            _LOG.info("The value is 'ALL' and 'allowed_all=True', returning..")
            return value.lower()
        _LOG.info(f"Checking whether {value} is a valid region...")
        if value not in ALL_REGIONS:
            _LOG.error(f"Invalid region '{value}' was given")
            self.fail(f"Value '{value}' is not a valid region. Try one of "
                      f"these: {ALL_REGIONS}", param, ctx)
        _LOG.info(f"Value '{value}' is a valid region, returning..")
        return value

    def get_metavar(self, param):
        shorten_regions = [ALL_REGIONS[0], "...", ALL_REGIONS[-1]]
        if self.allowed_all:
            shorten_regions.insert(0, self.ALL_VALUE)
        return f"[{'|'.join(shorten_regions)}]"


class DictParamType(click.types.StringParamType):
    name = 'dict'
    ITEMS_SEPARATOR = ','
    KEY_VALUE_SEPARATOR = ':'

    def convert(self, value, param, ctx):
        value = super().convert(value, param, ctx)
        _LOG.info(f'Stripping {value} from "{self.ITEMS_SEPARATOR}" a bit..')
        value = value[1:] if value.startswith(self.ITEMS_SEPARATOR) else value
        value = value[:-1] if value.endswith(self.ITEMS_SEPARATOR) else value
        result = {}
        _LOG.info(f'Converting: {value} to dict..')

        try:
            for item in value.split(self.ITEMS_SEPARATOR):
                k, v = item.split(self.KEY_VALUE_SEPARATOR)
                result[k.lstrip().rstrip()] = v.lstrip().rstrip()
        except ValueError as e:
            raise BadParameter(
                f'Wrong format: "{value}". '
                f'Must be "key:value" or "key1:value1,key2:value2". '
                f'\nError: {e.__str__()}')

        _LOG.info(f'Converted to such a dict: {result}')
        return result

    def get_metavar(self, param):
        return f'KEY1{self.KEY_VALUE_SEPARATOR}VALUE1' \
               f'{self.ITEMS_SEPARATOR}KEY2{self.KEY_VALUE_SEPARATOR}VALUE2'


class DeepDictParamType(click.types.StringParamType):
    name = 'deep-dict'
    ITEMS_SEPARATOR = ','
    SUB_KEY_VALUE_SEPARATOR = ':'
    MAIN_KEY_VALUE_SEPARATOR = ';'

    def convert(
            self,
            value: str,
            param: click.Parameter | None,
            ctx: click.Context | None,
    ) -> dict[str, dict[str, str]]:
        """
        Convert a string representation of nested key-value pairs into a
        dictionary of dictionaries

        **Example**::
            Input: 'volume;Name:DBStorage01,Environment:Prod'
            Output: {'volume': {'Name': 'DBStorage01', 'Environment': 'Prod'}}
        """
        value = super().convert(value, param, ctx)
        result = {}

        try:
            main_parts = value.split(self.MAIN_KEY_VALUE_SEPARATOR)
            if len(main_parts) != 2:
                raise InvalidValueError(
                    "Expected exactly one main key-value separator (';')"
                )
            main_key = main_parts[0].strip()
            sub_items = main_parts[1].strip()

            sub_dict = {}
            for item in sub_items.split(self.ITEMS_SEPARATOR):
                sub_parts = item.split(self.SUB_KEY_VALUE_SEPARATOR)
                if len(sub_parts) != 2:
                    raise InvalidValueError(
                        "Expected exactly one sub key-value separator (':')"
                    )
                sub_key = sub_parts[0].strip()
                sub_value = sub_parts[1].strip()
                sub_dict[sub_key] = sub_value

            result[main_key] = sub_dict
        except (ValueError, InvalidValueError) as e:
            raise BadParameter(
                f'Wrong format: "{value}". Expected format is: '
                f'"main_key1;sub_key1:val1,sub_key2:val2" or similar. '
                f'\nError: {str(e)}'
            )
        return result

    def get_metavar(self, param) -> str:
        return (
            f'MAIN_KEY1{self.MAIN_KEY_VALUE_SEPARATOR}'
            f'SUB_KEY1{self.SUB_KEY_VALUE_SEPARATOR}VALUE1{self.ITEMS_SEPARATOR}'
            f'SUB_KEY2{self.SUB_KEY_VALUE_SEPARATOR}VALUE2'
        )


def validate_bucket_name(ctx, param, value):
    try:
        from syndicate.core.resources.s3_resource import validate_bucket_name
        bucket_name = value
        if '/' in value:
            bucket_name = value.split('/', 1)[0]
        validate_bucket_name(bucket_name)
        return value
    except (ValueError, InvalidValueError) as e:
        raise BadParameter(e.__str__())


def check_prefix(ctx, param, value):
    extended_prefix = ctx.params.get('extended_prefix')
    if value:
        value = value.lower().strip()
        if extended_prefix:
            result = \
                ConfigValidator.validate_extended_prefix(
                    param.human_readable_name, value)
        else:
            result = \
                ConfigValidator.validate_prefix_suffix(
                    param.human_readable_name, value)
        if result:
            raise BadParameter(result)
        return value


def check_suffix(ctx, param, value):
    if value:
        value = value.lower().strip()
        result = \
            ConfigValidator.validate_prefix_suffix(param.human_readable_name,
                                                   value)
        if result:
            raise BadParameter(result)
        return value


def resolve_project_path(ctx, param, value):
    from syndicate.core import CONFIG
    if not value:
        USER_LOG.info(
            f"Parameter: '{param.human_readable_name}' wasn't specified. "
            f"Getting automatically")
        value = CONFIG.project_path \
            if CONFIG and CONFIG.project_path else os.getcwd()
        USER_LOG.info(f"Path: '{value}' was assigned to the "
                      f"parameter: '{param.human_readable_name}'")
    return value


def check_lambda_name(value):
    """Validates lambda's name"""
    _LOG.info(f"Validating lambda name: '{value}'")
    invalid_character = re.search('[^0-9a-zA-Z\-\_]', value)
    error = None
    if not 3 <= len(value) <= 63:
        error = f'lambda name \'{value}\' length must be between 3 and 63 characters'
    elif invalid_character:
        error = f'lambda name \'{value}\' contains invalid characters: ' \
                f'{invalid_character.group()}'
    elif value.startswith('-'):
        error = f"lambda name '{value}' cannot start with '-'"
    elif value.endswith('-'):
        error = f"lambda name '{value}' cannot end with '-'"
    if error:
        _LOG.error(f"Lambda name validation error: {error}")
        raise InvalidValueError(error)
    _LOG.info(f"Lambda name: '{value}' passed the validation")


def check_lambdas_names(ctx, param, value):
    """Applies lambda name validator for each lambda's name"""
    for lambda_name in value:
        try:
            check_lambda_name(lambda_name)
        except InvalidValueError as e:
            raise click.BadParameter(e.__str__(), ctx, param)
    return value


def check_lambda_layer_name(ctr, param, value):
    pattern = r'^[a-zA-Z0-9][a-zA-Z0-9-_]{0,63}$'
    errors = []
    if len(value) > 64:
        errors.append('The length of lambda layer name must be less or equal '
                      'to 64 character')
    if not value[0].isalpha():
        errors.append('The first character of the lambda layer name must be '
                      'a letter')
    if not re.match(pattern, value):
        errors.append('The lambda layer name must contain only lowercase '
                      'letters, numbers, underscores and hyphens')
    if errors:
        raise BadParameter(f'The lambda layer name is invalid. Details:\n'
                           f'{errors}')
    return value


def check_lambda_existence(ctr, param, value):
    from syndicate.core import PROJECT_STATE
    lambdas = PROJECT_STATE.lambdas
    for lambda_name in value:
        if lambda_name not in lambdas:
            raise BadParameter(
                f'Lambda with name \'{lambda_name}\' not found. '
                f'Please check the lambda name and try again')
    return value


def handle_interruption(_num: SIGINT, _frame):
    """ Meant to handle interruption signal, by releasing any given lock """
    _naming, _lock_types = 'PROJECT_STATE', (MODIFICATION_LOCK, WARMUP_LOCK)
    if _num == SIGINT:
        from syndicate.core import PROJECT_STATE
        _state = PROJECT_STATE
        if isinstance(_state, ProjectState):
            _locked_type = next((each for each in _lock_types
                                 if not _state.is_lock_free(each)), None)
            if _locked_type:
                _LOG.warn(f'Releasing the project state lock {_locked_type},'
                          'due to user interruption.')
                _state.release_lock(_locked_type)
                sync_project_state()
    sys.exit(_num)


def check_lambda_state_consistency(objected_lambdas: list,
                                   subjected_lambdas: dict, runtime: str):
    from syndicate.core.groups import RUNTIME
    return next((True for each in objected_lambdas if subjected_lambdas.get(
        each, {}).get(RUNTIME) == runtime), False)


def delete_none(_dict):
    """Delete None values recursively"""
    if isinstance(_dict, dict):
        for key, value in list(_dict.items()):
            if isinstance(value, (list, dict, tuple, set)):
                _dict[key] = delete_none(value)
            elif value is None or key is None:
                del _dict[key]

    elif isinstance(_dict, (list, set, tuple)):
        _dict = type(_dict)(
            delete_none(item) for item in _dict if item is not None)
    return _dict


def without_zip_ext(name: str) -> str:
    _zip = '.zip'
    if name.endswith(_zip):
        name = name[:-len(_zip)]
    return name


def zip_ext(name: str) -> str:
    _zip = '.zip'
    if not name.endswith(_zip):
        name = name + _zip
    return name


def is_zip_empty(zip_path):
    """Check if a ZIP file is empty"""
    try:
        with zipfile.ZipFile(zip_path, 'r') as zip_file:
            contents = zip_file.namelist()
            if not contents:
                return True
            else:
                return False
    except zipfile.BadZipFile:
        _LOG.info('The file is not a zip file or it is corrupted.')
        return True
    except FileNotFoundError:
        _LOG.info('The zip file does not exist.')
        return True


def check_file_extension(ctx, param, value, extensions):
    for extension in extensions:
        if value is None:
            return
        if value.lower().endswith(extension):
            return value
    raise BadParameter(f'Only files with extensions {extensions} are '
                       f'supported.')


def validate_incompatible_options(ctx, param, value, incompatible_options):
    if value:
        conflict_options = [
            option for option in incompatible_options if
            ctx.params.get(option.replace('-', '_')) or
            ctx.params.get(option)
        ]
        if conflict_options:
            raise BadParameter(
                f'Parameter \'{param.human_readable_name}\' is incompatible '
                f'with {conflict_options}'
            )
        return value


def validate_authorizer_name_option(ctx, param, value):
    if value:
        authorization_type = ctx.params.get('authorization_type')
        if not authorization_type:
            raise BadParameter(
                f'Parameter \'{param.human_readable_name}\' can\'t be used '
                f'without \'authorization-type\' parameter')
        if authorization_type != CUSTOM_AUTHORIZER_KEY:
            raise BadParameter(
                f"Parameter '{param.human_readable_name}' can't be used "
                f"with 'authorization_type' '{authorization_type}")
        return value


def validate_api_gw_path(ctx, param, value):
    pattern = (
        r'^/(?:([a-zA-Z0-9-._~]+|\{[a-zA-Z0-9-._~]+\})/)*([a-zA-Z0-9-._~]+|'
        r'\{[a-zA-Z0-9-._~]+\}|\{proxy\+\})$')
    _LOG.debug(
        f"The parameter '--{param.human_readable_name}' value is '{value}'")
    if os.name == 'nt' and Path(value).is_absolute():
        raise BadParameter(
            f"Your terminal resolves the parameter "
            f"'--{param.human_readable_name}' value as a filesystem path. "
            f"Please pass the parameter '--{param.human_readable_name}' "
            f"value without starting slash ('/')")
    if not value.startswith('/'):
        value = '/' + value
    if not re.match(pattern, value):
        raise BadParameter(
            f"'{value}'. "
            f"A valid API gateway path must begin with a '/' and can contain "
            f"alphanumeric characters, hyphens, periods, underscores or "
            "dynamic parameters wrapped in '{}'. "
        )
    return value


def check_tags(ctx, param, value):
    if value:
        errors = validate_tags(param.human_readable_name, value)
        if errors:
            raise BadParameter(errors)
        return value


def validate_tags(key_name, tags_dict):
    return ConfigValidator.validate_tags(key_name, tags_dict)


def set_debug_log_level(ctx, param, value):
    if value:
        loggers = [logging.getLogger(name) for name in
                   logging.root.manager.loggerDict if
                   name.startswith(LOG_NAME) or
                   name.startswith(USER_LOG_NAME)]

        console_handler = logging.getLogger(USER_LOG_NAME).handlers[0]

        for logger in loggers:
            if not logger.isEnabledFor(logging.DEBUG):
                logger.setLevel(logging.DEBUG)
                if logger.name == LOG_NAME:
                    logger.addHandler(console_handler)
        _LOG.debug('The logs level was set to DEBUG')


def verbose_option(func):
    @click.option('--verbose', '-v', is_flag=True,
                  callback=set_debug_log_level, expose_value=False,
                  is_eager=True, help="Enable logging verbose mode.")
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def compute_file_hash(file_path: Union[str, Path],
                      algorithm: str = 'sha256') -> str:
    hash_obj = hashlib.new(algorithm)
    with open(file_path, 'rb') as f:
        while True:
            chunk = f.read(4096)
            if not chunk:
                break
            hash_obj.update(chunk)
    return hash_obj.hexdigest()


def resolve_deploy_target_bucket_param(ctx, param, value):
    if bundle_bucket_name := ctx.params.get('bundle_bucket_name'):
        USER_LOG.warn(
            "The parameter '--bundle-bucket-name' is deprecated! "
            "It is highly recommended to use '--deploy-target-bucket' instead."
        )

    bucket_name = value or bundle_bucket_name
    if not bucket_name:
        raise MissingParameter()

    return validate_bucket_name(ctx, param, bucket_name)


def are_resource_types_valid(param_name: str,
                             types: list[str] | None,
                             allowed_types: list[str]) -> bool:
    """
    Validate incoming from click AWS resource types
    """
    if not types:
        return True

    invalid_types = [t for t in types if
                     t not in allowed_types]
    if invalid_types:
        USER_LOG.error(
            f"Invalid resource type(s) in `{param_name}` "
            f"parameter: {', '.join(invalid_types)}. "
            f"Allowed types: {', '.join(allowed_types)}"
        )
        return False
    return True


def strip_prefix_suffix(res_name: str) -> str:
    """
    Strips the resource prefix and suffix if it is present.
    """
    from syndicate.core import CONFIG
    if CONFIG.resources_prefix and res_name.startswith(CONFIG.resources_prefix):
        res_name = res_name[len(CONFIG.resources_prefix):]
    if CONFIG.resources_suffix and res_name.endswith(CONFIG.resources_suffix):
        res_name = res_name[:-len(CONFIG.resources_suffix)]
    return res_name
